var LinkEnhancr = {
    URL_REGEX: /(http(s?):\/\/($|[^ ]+)|www\.*.?\.($|[^ ]+))/i,
    PASTE_REGEX: /^http(s?):\/\/[.*\S^$]+$/i,
    NODES_POOL: {},
    CARD_DATA: {},
    LINK_ENHANCR_ATTACHMENT_TYPE_INLINE: "LINK_ENHANCR_ATTACHMENT_TYPE_INLINE",
    LINK_ENHANCR_ATTACHMENT_TYPE_OUTLINE: "LINK_ENHANCR_ATTACHMENT_TYPE_OUTLINE",
    needsReturn: true,

    // MARK: - Public methods
    // All methods below must return a value to avoid `Fatal error: Bug in WebKit: Received neither result or failure` when evaluated via `WKWebView.evaluateJavaScript()`
    enhanceLinks: function(contentEditable) {
        var _this = this,
            nodes = {},
            hasLinks = this._surroundInElement(contentEditable, nodes),
            urls = new Array();

        if (hasLinks && linkEnhancrEnabled) {
            Object.keys(nodes).forEach(function (key) {
                urls.push("url=" + encodeURIComponent(key));
                _this.NODES_POOL[key] = nodes[key];
            });
            this._fetchCardsForURLs(urls);
        }
        InputHandler.scrollIfNeeded();
        
        return true;
    },

    handlePasteEvent: function(event) {
        if (!/text\/plain/.test(event.clipboardData.types) || !this.PASTE_REGEX.test(event.clipboardData.getData('text/plain'))) {
            return false;
        }
        event.preventDefault();
        this.needsReturn = false;

        var urlString = event.clipboardData.getData('text/plain'),
            linkNode = this._createLinkElement(urlString),
            textNode = document.createTextNode("\u00A0"),
            placeholderNode;

        window.getSelection().getRangeAt(0).insertNode(linkNode);

        linkNode.parentNode.insertBefore(textNode, linkNode.nextSibling);
        this._setCursorPositionAtNode(textNode, 1);

        if (linkEnhancrEnabled) {
            placeholderNode = this._createCardPlaceholderElement("temp", "medium");
            linkNode.parentNode.insertBefore(placeholderNode, textNode.nextSibling);

            this.NODES_POOL[urlString] = {
                linkNode: linkNode,
                placeholderNode: placeholderNode
            };

            this._fetchCardsForURLs(["url=" + encodeURIComponent(urlString)]);
        }
        
        return true;
    },

    createCard: function(urlString, cardData) {
        var node = this.NODES_POOL[urlString],
            linkNode = node["linkNode"],
            placeholderNode = node["placeholderNode"],
            parentNode = placeholderNode.parentNode,
            lineBreakBeforeCard = this._createLineBreakElement(),
            lineBreakAfterCard = this._createLineBreakElement(),
            cardElement;

        this.CARD_DATA[cardData["cardId"]] = cardData;

        linkNode.innerHTML = cardData["cardTitle"];
        linkNode.setAttribute("href", cardData["url"]);
        linkNode.setAttribute("class", cardData["cardId"]);

        cardElement = this._getCardElement(cardData, null);

        parentNode.insertBefore(lineBreakBeforeCard, placeholderNode);
        parentNode.insertBefore(lineBreakAfterCard, placeholderNode.nextSibling);
        parentNode.replaceChild(cardElement, placeholderNode);

        this._setCursorPositionAtNode(lineBreakAfterCard.nextSibling, 0);

        InputHandler.scrollIfNeeded();
        delete this.NODES_POOL[urlString];
        
        return true;
    },

    removeCard: function(cardId, convertLink) {
        var cardElement = document.querySelector("[id*='" + cardId + "']:not([id*=placeholder])"),
            previewsContainer = document.querySelector("[class*='enhancr-link-previews']"),
            linkNode = document.querySelector("a[class*='" + cardId + "']");
        if (linkNode && convertLink) {
            linkNode.innerHTML = linkNode.getAttribute("href");
        }

        cardElement.parentNode.removeChild(cardElement);

        if (previewsContainer && document.querySelectorAll("[class*='enhancr-link-previews'] > div").length == 0) {
            previewsContainer.parentNode.removeChild(previewsContainer);
        }
        
        return true;
    },

    moveCardToBottom: function(cardId) {
        if (this.CARD_DATA[cardId]) {
            this._moveCardToBottom(cardId, this.CARD_DATA[cardId]);
        } else {
            window.location = "ios://fetch-link-info/?url=" + encodeURIComponent(this.getURLForCardId(cardId)) + "&move=bottom&cardid=" + cardId;
        }
        
        return true;
    },

    moveCardInline: function(cardId) {
        if (this.CARD_DATA[cardId]) {
            this._moveCardInline(cardId, this.CARD_DATA[cardId]);
        } else {
            window.location = "ios://fetch-link-info/?url=" + encodeURIComponent(this.getURLForCardId(cardId)) + "&move=inline&cardid=" + cardId;
        }
        
        return true;
    },

    moveFetchedCardToBottom: function(cardId, urlString, cardData) {
        this.CARD_DATA[cardData["cardId"]] = cardData;
        this._moveCardToBottom(cardId, cardData);
        
        return true;
    },

    moveFetchedCardInline: function(cardId, urlString, cardData) {
        this.CARD_DATA[cardData["cardId"]] = cardData;
        this._moveCardInline(cardId, cardData);
        
        return true;
    },

    prepareLinkEnhancrCards: function() {
        var moreButton,
            elements = document.querySelectorAll("[class*='yahoo-link-enhancr-card']");
        for (i in elements) {
            var cardElement = elements[i];
            if (!(cardElement instanceof HTMLElement)) {
                continue;
            }
            if (cardElement.querySelector("[class*='card-primary-image-cell']") == null && cardElement.className.indexOf("yahoo-link-enhancr-no-image") === -1) {
                cardElement.classList.add("yahoo-link-enhancr-no-image");
            }
            if (cardElement.querySelector("button.more") == null && document.querySelector("blockquote #" + cardElement.getAttribute("id")) == null) {
                moreButton = this._getAppendMoreButton(cardElement.getAttribute("id"));
                cardElement.appendChild(moreButton);
                if (cardElement.getAttribute("data-size") == "append") {
                    cardElement.setAttribute("data-type", this.LINK_ENHANCR_ATTACHMENT_TYPE_OUTLINE);
                } else {
                    cardElement.setAttribute("data-type", this.LINK_ENHANCR_ATTACHMENT_TYPE_INLINE);
                }
            }
        }
        
        return true;
    },

    prepareLinkEnhancrCardsForSharing: function() {
        var shareButton,
            elements = document.querySelectorAll("[class*='yahoo-link-enhancr-card']");
        for (i in elements) {
            var cardElement = elements[i];
            if (!(cardElement instanceof HTMLElement)) {
                continue;
            }
            if (cardElement.querySelector("[class*='card-primary-image-cell']") == null && cardElement.className.indexOf("yahoo-link-enhancr-no-image") === -1) {
                cardElement.classList.add("yahoo-link-enhancr-no-image");
            }
            if (cardElement.querySelector("button.share") == null) {
                shareButton = this._getShareButton(cardElement.getAttribute("id"));
                cardElement.appendChild(shareButton);
                if (cardElement.getAttribute("data-size") == "append") {
                    cardElement.setAttribute("data-type", this.LINK_ENHANCR_ATTACHMENT_TYPE_OUTLINE);
                } else {
                    cardElement.setAttribute("data-type", this.LINK_ENHANCR_ATTACHMENT_TYPE_INLINE);
                }
            }
        }
        
        return true;
    },

    getURLForCardId: function(cardId) {
        return document.querySelector("[id*='" + cardId + "']:not([id*=placeholder]) > [class*='yahoo-enhancr-cardlink']").getAttribute("href");
    },

    // MARK: Private methods

    _fetchCardsForURLs: function(urls) {
        // Schedule fetch to the next event
        setTimeout(function() {
            window.location = "ios://fetch-link-info/?" + urls.join("&");
        }, 100);
    },

    _getAppendMoreButton: function(cardId) {
        var moreButton = document.createElement("button");
        moreButton.className = "more";
        moreButton.contentEditable = false;
        moreButton.setAttribute("data-guid", cardId);
        moreButton.setAttribute("data-action", "moreMenuInlineAttachment");
        moreButton.name = "showInlineImageMenu";
        moreButton.setAttribute("aria-label", LocalizedString.attachment_more_menu);
        return moreButton;
    },

    _getShareButton: function(cardId) {
        var button = document.createElement("button");
        button.className = "share";
        button.contentEditable = false;
        button.setAttribute("data-guid", cardId);
        button.setAttribute("data-action", "showShareMenu");
        button.name = "showShareMenu";
        button.setAttribute("aria-label", "share menu");
        return button;
    },

    _setCursorPositionAtNode: function(node, cursorOffset) {
        var range = document.createRange(),
            selection = window.getSelection();

        if (node) {
            range.setStart(node, cursorOffset);
            range.setEnd(node, cursorOffset);

            selection.removeAllRanges();
            selection.addRange(range);
        }

        setTimeout(function() {
            InputHandler.scrollIfNeeded();
        }, 100);
    },

    _moveCardToBottom: function(cardId, cardData) {
        var firstBlockquote = document.querySelector("[class*='yahoo-quoted-begin']"),
            cardElement = document.querySelector("[id*='" + cardId + "']:not([id*=placeholder])"),
            placeholderElement = this._createCardPlaceholderElement(cardId, cardElement.getAttribute("data-size")),
            previewsContainer = document.querySelector("[class*='enhancr-link-previews']");

        cardElement.parentNode.replaceChild(placeholderElement, cardElement);

        cardData["cardId"] = cardId;
        cardElement = this._createCardElement("enhancr_card_append", cardData, true);

        if (!previewsContainer) {
            previewsContainer = this._createCardPreviewsElement(LocalizedString.links_in_the_message);
            if (firstBlockquote) {
                firstBlockquote.parentNode.insertBefore(previewsContainer, firstBlockquote);
            } else {
                MessageComposeController.editableElement().appendChild(previewsContainer);
            }
        }
        previewsContainer.appendChild(cardElement);
    },

    _moveCardInline: function(cardId, cardData) {
        var linkNode = document.querySelector("[class*='" + cardId + "']"),
            placeholderNode = document.querySelector("[id*='" + cardId + "-enhancr-placeholder']"),
            parentNode,
            nodeToInsertAfter,
            template = "enhancr_card_medium";

        if (placeholderNode) {
            parentNode = placeholderNode.parentNode;
            nodeToInsertAfter = placeholderNode;
            if (placeholderNode.getAttribute("data-size") == "mini") {
                template = "enhancr_card_mini";
            } else if (placeholderNode.getAttribute("data-size") == "cover") {
                template = "enhancr_card_cover";
            }
        } else if (linkNode) {
            parentNode = linkNode.parentNode;
            nodeToInsertAfter = linkNode;
        } else {
            parentNode = MessageComposeController.editableElement();
            nodeToInsertAfter = parentNode.firstChild;
        }

        this.removeCard(cardId, false);
        cardData["cardId"] = cardId;

        parentNode.insertBefore(this._getCardElement(cardData, template), nodeToInsertAfter.nextSibling, false);
        if (placeholderNode) {
            parentNode.removeChild(placeholderNode);
        } else {
            parentNode.insertBefore(function() {
                var div = document.createElement("div");
                div.appendChild(document.createElement("br"));
                return div;
            }(), nodeToInsertAfter.nextSibling);
        }
        InputHandler.scrollIfNeeded();
    },

    _createLinkElement: function(urlString) {
        var linkNode = document.createElement("a");

        linkNode.href = urlString;
        linkNode.target = "_blank";
        linkNode.innerHTML = urlString;

        return linkNode;
    },

    _createLineBreakElement: function() {
        var div = document.createElement("div");
        div.appendChild(document.createElement("br"));
        return div;
    },

    _getCardElement: function(cardData, template) {
        if (!template) {
            template = "enhancr_card_medium";
        }

        if (!cardData["primaryImage"] && template == "enhancr_card_medium") {
            template = "enhancr_card_medium_no_image";
        }
        return this._createCardElement(template, cardData, false);
    },

    _createCardElement: function(template, cardData, fullyTappable) {
        var cardElement = Template.loadTemplate(template, cardData),
            moreButton = this._getAppendMoreButton(cardData["cardId"]);

        cardElement.appendChild(moreButton);

        cardElement.setAttribute("data-type", (cardElement.getAttribute("data-size") == "append") ? this.LINK_ENHANCR_ATTACHMENT_TYPE_OUTLINE: this.LINK_ENHANCR_ATTACHMENT_TYPE_INLINE);

        if (cardElement.querySelector("[class*='card-primary-image-cell']") == null) {
            cardElement.classList.add("yahoo-link-enhancr-no-image");
        }

        return cardElement;
    },

    _createCardPlaceholderElement: function(cardId, cardSize) {
        return Template.loadTemplate("enhancr_placeholder", {
            cardId: cardId,
            cardSize: cardSize
        });
    },

    _createCardPreviewsElement: function(title) {
        return Template.loadTemplate("enhancr_previews", {
            title: title
        });
    },

    _shouldLinkifyContents: function(el) {
        return el.tagName != "A";
    },

    _surroundInElement: function(el, nodes) {
        var child = el.lastChild,
            hasLinks = false;

        while (child) {
            if (child.nodeType == 1 && this._shouldLinkifyContents(el)) {
                if (this._surroundInElement(child, nodes)) {
                    hasLinks = true;
                }
            } else if (child.nodeType == 3) {
                if (this._surroundMatchingText(child, nodes)) {
                    hasLinks = true
                }
            }
            child = child.previousSibling;
        }
        return hasLinks;
    },

    _surroundMatchingText: function(textNode, nodes) {
        var hasLinks = false,
            parent = textNode.parentNode,
            result, surroundingNode, matchedTextNode, matchLength, matchedText, url,

            textNode, placeholderNode;

        while (textNode && (result = this.URL_REGEX.exec(textNode.data))) {
            matchedTextNode = textNode.splitText(result.index);
            if (!this._shouldLinkifyContents(matchedTextNode.parentNode)) {
                continue;
            }
            matchedText = result[0];
            matchLength = matchedText.length;
            textNode = (matchedTextNode.length > matchLength) ?
            matchedTextNode.splitText(matchLength) : null;
            if (this._elementHasParentOfType(matchedTextNode, "blockquote")) {
                continue;
            }
            surroundingNode = this._createLinkElement(this._prependProtocolIfNeeded(matchedTextNode.cloneNode(true).data.trim()));
            url = this._prependProtocolIfNeeded(matchedTextNode.data.trim());
            parent.replaceChild(surroundingNode, matchedTextNode);
            if (linkEnhancrEnabled) {
                placeholderNode = this._createCardPlaceholderElement("temp", "medium");
                parent.insertBefore(placeholderNode, surroundingNode.nextSibling);
            }
            if (!this.needsReturn) {
                textNode = document.createTextNode("\u00A0");
                parent.insertBefore(textNode, surroundingNode.nextSibling);
                this._setCursorPositionAtNode(textNode, 1);
            }
            nodes[url] = {
                linkNode: surroundingNode,
                placeholderNode: placeholderNode
            };
            hasLinks = true;
        }
        return hasLinks;
    },

    _elementHasParentOfType: function(element, tagName) {
        var currentNode = element.parentNode;
        while (currentNode) {
            if (currentNode.tagName && currentNode.tagName.toLowerCase() === tagName) {
                return true;
            }
            currentNode = currentNode.parentNode;
        }
        return false;
    },

    _prependProtocolIfNeeded: function(urlString) {
        urlString = urlString[0].toLowerCase() + urlString.substr(1);
        return urlString.toLowerCase().indexOf("http") == 0 ? urlString : "http://" + urlString;
    }
}
