//Wed Jan  3 16:50:06 PST 2018
if (typeof(Stationery) == "undefined") {
    Stationery = {};
}

Stationery.constants = {
    mainComposeAreaID: (typeof(COMPOSE_CONTENT_ELEMENT_ID) !== 'undefined') ? COMPOSE_CONTENT_ELEMENT_ID : 'yahoo__compose_area',
    paperlessPostThemeProviderName: 'Paperless Post',
    motifTemplateName: 'motif_template',
    rteStationeryCoreClassName: 'rte-stationery-core',
    mobstoreBaseImageSrc: 'https://s.yimg.com/nq/tictac/stationery/themes/',
    clearStationeryClassName: 'btn-clear-stationery',
    stationeryEditableBodyID: 'middleBody',
    yahooQuotedBeginQuerySelector: '[class*=yahoo-quoted-begin]',
    outlineEnhancrQuerySelector: '[class*=enhancr-link-previews]',
    tableWrapperSelector: 'table[class*=wrapper]',
    noneStationeryThemeId: 'clearStationeryID',
    coreQuotedContentQuerySelector: 'blockquote',
    unoTemplateName: 'uno_template',
    dayms: 24 * 60 * 60 * 1000
};

Stationery.dateUtils = {
    getMsFromDateString: function(dateString) {
        var dateInfo = dateString.split('-').map(num => parseInt(num));

        if (dateInfo.length < 3 || isNaN(dateInfo[0]) || isNaN(dateInfo[1]) || isNaN(dateInfo[2])) {
            return 0;
        }

        return (new Date(dateString)).getTime();
    },
    dateRangeContainsNow: function(dateInfoArray) {
        if (dateInfoArray.length < 2) {
            return false;
        }
        var start = dateInfoArray[0];
        var end = dateInfoArray[1];
        var now = Date.now();
        
        return this.getMsFromDateString(start) <= now && now < this.getMsFromDateString(end) + Stationery.constants.dayms;
    },
    dateRangeStartsBeforeNow: function(arr) {
        // arr[0] = start date
        return this.getMsFromDateString(arr[0]) <= Date.now();
    },
    sortConfigsByStartDate: function(a, b) {
        var startA = a.dates[0];
        var endA = a.dates[1];
        var startB = b.dates[0];
        var endB = b.dates[1];

        // if the start dates are the same, sort the later end date last
        return (
            this.getMsFromDateString(startB) - this.getMsFromDateString(startA) ||
            this.getMsFromDateString(endA) - this.getMsFromDateString(endB)
        );
    },
    sortConfigsByEndDate: function(a, b) {
        var startA = a.dates[0];
        var endA = a.dates[1];
        var startB = b.dates[0];
        var endB = b.dates[1];

        // if the end dates are the same, sort the later start date first
        return (
            this.getMsFromDateString(endB) - this.getMsFromDateString(endA) ||
            this.getMsFromDateString(startB) - this.getMsFromDateString(startA)
        );
    }
}

Stationery.getLocalizedString = function(strings) {
    var lang = (Stationery.deviceMetrics && Stationery.deviceMetrics.lang) || "en",
        interimLang = lang.substring(0, lang.lastIndexOf("-")),
        shortLang = lang.split("-")[0];
    return strings[lang] || strings[interimLang] || strings[shortLang] || strings["en"];
};

Stationery.getCoreBodyWithTheme = function(themeId) {
    var themeObj = JSON.parse(JSON.stringify(this.stationeryThemeConfig.stationeryThemes[themeId])),
        themeVersion = JSON.parse(JSON.stringify(this.stationeryThemeConfig.themeVersions[themeId])),
        templateName = themeObj.templateName,
        mobstoreBaseUrl = this.stationeryThemeConfig.stationeryMobThmPath,
        blockQuoteNode, yahooQuotedNode, coreContent, stationeryBodyNode, outlineEnhancrNode,
        rteCloneNode = document.createElement('div');
    
    if (mobstoreBaseUrl.charAt(mobstoreBaseUrl.length - 1) !== '/') {
        mobstoreBaseUrl += '/';
    }

    rteCloneNode.innerHTML = document.querySelector('[id*=' + Stationery.constants.mainComposeAreaID + ']').innerHTML;
    blockQuoteNode = rteCloneNode.querySelector(Stationery.constants.coreQuotedContentQuerySelector);
    outlineEnhancrNode = rteCloneNode.querySelector(Stationery.constants.outlineEnhancrQuerySelector);

    if (blockQuoteNode) {
        blockQuoteNode.remove();
    }

    yahooQuotedNode = rteCloneNode.querySelector(Stationery.constants.yahooQuotedBeginQuerySelector);
    if (yahooQuotedNode) {
        yahooQuotedNode.remove();
    }

    if (outlineEnhancrNode) {
        outlineEnhancrNode.remove();
    }

    stationeryBodyNode = rteCloneNode.querySelector('[id*=' + Stationery.constants.stationeryEditableBodyID + ']');
    if (stationeryBodyNode) {
        coreContent = stationeryBodyNode.innerHTML;
    } else {
        coreContent = rteCloneNode.innerHTML;
    }

    themeObj.yahooStationery = true;
    themeObj.pplogo = themeObj.themeProviderName === "Paperless Post";
    if (typeof(LocalizedString) === 'object') {
        themeObj.msg_footer = LocalizedString.stationery_yahoo_attribution_footer;
    } else {
        themeObj.msg_footer = Stationery.strings ? Stationery.getLocalizedString(Stationery.strings.str_yahoo_mail_stationery) : "Yahoo Mail Stationery";
    }
    if (themeObj.topImg_src) {
        themeObj.topImg_src = mobstoreBaseUrl + themeVersion + '/' + themeObj.uniqueThemeName + '/' + themeObj.topImg_src;
    }
    if (themeObj.bottomImg_src) {
        themeObj.bottomImg_src = mobstoreBaseUrl + themeVersion + '/' + themeObj.uniqueThemeName + '/' + themeObj.bottomImg_src;
    }
    themeObj.messageText = coreContent;
    return Mustache.render(this.templates[templateName], themeObj);

};

Stationery.getReplyForwardBody = function() {
    var rteComposeNode = document.querySelector('[id*=' + Stationery.constants.mainComposeAreaID + ']'),
        coreContent = '',
        yahooQuotedNode = rteComposeNode.querySelector(Stationery.constants.yahooQuotedBeginQuerySelector),
        replyForwardQuotedNode = rteComposeNode.querySelector(Stationery.constants.coreQuotedContentQuerySelector),
        outlineEnhancrNode = rteComposeNode.querySelector(Stationery.constants.outlineEnhancrQuerySelector);

    if (yahooQuotedNode) {
        coreContent += yahooQuotedNode.outerHTML;
    }

    if (replyForwardQuotedNode) {
        coreContent += replyForwardQuotedNode.outerHTML;
    }

    if (outlineEnhancrNode) {
        coreContent += outlineEnhancrNode.outerHTML;
    }

    return coreContent;
};

Stationery.applyStationery = function(themeId) {
    // construct the new Compose Body
    var _this = this,
        replyForwardBody = _this.getReplyForwardBody(),
        coreBody = _this.getCoreBodyWithTheme(themeId),
        clearStationeryButton,
        rteStationeryNode, rteComposeNode = document.querySelector('[id*=' + Stationery.constants.mainComposeAreaID + ']');

    rteComposeNode.innerHTML = coreBody + replyForwardBody;

    // Attach the handlers and contentEditableFalse
    clearStationeryButton = document.querySelector('[class*=' + Stationery.constants.clearStationeryClassName + ']');
    clearStationeryButton.addEventListener('click', function() {
        _this.clearStationery();
    });

    rteComposeNode.contentEditable = 'false';
    rteStationeryNode = document.querySelector('[class*=' + Stationery.constants.rteStationeryCoreClassName + ']');
    rteStationeryNode.addEventListener('click', function(evt) {
        if (evt.target.hasAttribute('data-action')) {
            return;
        }
        var stationeryBodyNode = document.querySelector('[id*=' + Stationery.constants.stationeryEditableBodyID + ']');
        if (stationeryBodyNode) {
            stationeryBodyNode.focus();
        }
    }, false);

    _this.applyBehavior(themeId);

    //ios only
    if (window.bodyUpdated) {
        bodyUpdated();
    }

    // android only
    if (window.yMailBridge) {
        yMailBridge.setStationeryId(themeId);
    }
};

Stationery.initialize = function() {
    var rteStationeryNode, mainRTENode, stationeryBodyNode, clearStationeryButton, tableWrapperNode;
    mainRTENode = document.querySelector('[id*=' + Stationery.constants.mainComposeAreaID + ']');
    if (mainRTENode) {
        Stationery.appendStationeryStyles();
    }
    if (this.containsStationery()) {
        if (mainRTENode) {
            rteStationeryNode = document.querySelector('[class*=' + Stationery.constants.rteStationeryCoreClassName + ']');
            stationeryBodyNode = document.querySelector('[id*=' +  Stationery.constants.stationeryEditableBodyID + ']');
            rteStationeryNode.contentEditable = "false";
            mainRTENode.contentEditable = "false";
            stationeryBodyNode.contentEditable = "true";

            clearStationeryButton = rteStationeryNode.querySelector('.' + Stationery.constants.clearStationeryClassName);
            if (!clearStationeryButton) {
                clearStationeryButton = document.createElement('span');
                clearStationeryButton.className = Stationery.constants.clearStationeryClassName;
                tableWrapperNode = rteStationeryNode.querySelector(Stationery.constants.tableWrapperSelector);
                tableWrapperNode && tableWrapperNode.parentNode.insertBefore(clearStationeryButton, tableWrapperNode);
            }

            clearStationeryButton.addEventListener('click', function() {
                Stationery.clearStationery();
            });

            rteStationeryNode.addEventListener('click', function() {
                stationeryBodyNode.focus();
            });
        }

        var currentAppliedTheme = Stationery.currentAppliedTheme();
        Stationery.applyBehavior(currentAppliedTheme);
    }
};

Stationery.adjustToShowStationeryMode = function() {
    document.body.scrollTop = 0;
};

Stationery.appendStationeryStyles = function() {
    var str_clearStationery = Stationery.strings ? Stationery.getLocalizedString(Stationery.strings.str_clear_stationery) : "Clear Stationery";

    if (typeof(LocalizedString) === 'object') {
        str_clearStationery = LocalizedString.stationery_clear_stationery;
    }

    styleNode = document.createElement('style');
    styleNode.innerHTML = '.rte-stationery-core .btn-clear-stationery:after { content: "' + str_clearStationery + '"; }  blockquote .rte-stationery-core .btn-clear-stationery {display: none;} ';
    document.head.appendChild(styleNode);
};

Stationery.clearStationery = function() {
    var rteStationeryNode, mainRTENode, stationeryBodyNode, clearStationeryButton, coreHtmlContent;

    if (this.containsStationery()) {
        mainRTENode =  document.querySelector('[id*=' + Stationery.constants.mainComposeAreaID + ']');
        rteStationeryNode = document.querySelector('[class*=' + Stationery.constants.rteStationeryCoreClassName + ']');
        stationeryBodyNode = document.querySelector('[id*=' +  Stationery.constants.stationeryEditableBodyID + ']');

        rteStationeryNode.remove();
        coreHtmlContent = stationeryBodyNode.innerHTML;
        mainRTENode.innerHTML = coreHtmlContent + mainRTENode.innerHTML;
        mainRTENode.contentEditable = "true";

        this.clearBehavior();

        // ios only
        if (window.bodyUpdated) {
            bodyUpdated();
            setTimeout(function() {
                window.location = "ios://clear-stationery";
            }, 0);
        }

        // android only
        if (window.yMailBridge) {
            yMailBridge.setStationeryId(NONE_STATIONERY_ID);
            syncComposeContent(mainRTENode, false);
        }
    }
};

Stationery.currentAppliedTheme = function() {
    var imageNode, srcComponents;

    if (Stationery.containsStationery()) {
        imageNode = document.querySelector("img[src^='" + Stationery.constants.mobstoreBaseImageSrc + "']");
        srcComponents = imageNode.src.split("/");
        return srcComponents[srcComponents.length - 2];
    }
    return Stationery.constants.noneStationeryThemeId;
};

Stationery.containsStationery = function() {
    var mainRTENode = document.querySelector('[id*=' + Stationery.constants.mainComposeAreaID + ']') || document.body,
        totalStationeryCards, stationeryCardsInBlock, quotedNode;

    totalStationeryCards = mainRTENode.querySelectorAll("[class*='" + Stationery.constants.rteStationeryCoreClassName + "']").length;

    if (totalStationeryCards === 0) {
        return false;
    }

    quotedNode = mainRTENode.querySelector(Stationery.constants.coreQuotedContentQuerySelector);
    if (quotedNode) {
        stationeryCardsInBlock = quotedNode.querySelectorAll("[class*='" + Stationery.constants.rteStationeryCoreClassName + "']").length;
    } else {
        stationeryCardsInBlock = 0;
    }

    return (totalStationeryCards > stationeryCardsInBlock);
};

if (typeof(document) !== 'undefined') {
    document.addEventListener( "DOMContentLoaded", function() {
        Stationery.initialize();
    }, false );
}

Stationery.getThemesForToday = function(isCorp, lang) {
    var config = Stationery.stationeryThemeConfig;
    var dateUtils = Stationery.dateUtils;
    var themeDates = config.themeDates || [];
    var themeVersions = config.themeVersions;
    var stationeryThemes = config.stationeryThemes;
    var matchingConfigs = themeDates.filter(
        function(themeDate) {
            return dateUtils.dateRangeContainsNow(themeDate.dates);
        }).sort(dateUtils.sortConfigsByStartDate.bind(dateUtils));
    var categories = ['featured', 'everyday', 'birthday', 'seasonal'];

    // if there's no date range that contains the current date, then pick the
    // one that ended most recently but still started before today.  if that
    // list is empty, then the current date must be before we started using date
    // ranges, presumably because the user is a time traveler.  so just default
    // to the earliest date range config.
    var todaysThemes =
        matchingConfigs[0] ||
        themeDates.filter(
            function(themeDate) {
                return dateUtils.dateRangeStartsBeforeNow(themeDate.dates);
            }).sort(dateUtils.sortConfigsByEndDate.bind(dateUtils))[0] ||
        themeDates[0];
    var namedRanges = (todaysThemes && todaysThemes.dates[2]) || {};
    var themeIDs;

    if (todaysThemes) {
        // in case there's more than one category passed in, flatten the theme IDs
        // for all of the categories into one array, with no duplicates
        themeIDs = categories.reduce(
            function(array, category) {
                return array.concat(todaysThemes[category]);
            }, []);
    } else {
        // the themeDates array is empty for some reason, so just return all the
        // themes, which will be messy, but at least the picker won't be blank
        themeIDs = Object.keys(stationeryThemes);
    }
    
    var usedThemes = {};
    

    return themeIDs.reduce(function(themes, themeID) {
        var id = themeID;
        var range;

        if (themeID instanceof Array) {
            id = themeID[0];
            range = (themeID[1] instanceof Array) ? themeID[1] : namedRanges[themeID[1]];
        }

        if (usedThemes[id]) {
            return themes;
        }

        usedThemes[id] = true;
        
        var theme = stationeryThemes[id];
        var themeVersion = themeVersions[id];

        if (!theme || !themeVersion) {
            return themes;
        }

        var restricted = theme.restricted;
        var supportedLang_regex = theme.supportedLang_regex;

        // this duplicates the checks in getByCategories() below, but that code
        // will go away when we switch to always using date ranges
        if (
            (supportedLang_regex && lang && !supportedLang_regex.test(lang)) ||
            (restricted && !isCorp) ||
            (range && !dateUtils.dateRangeContainsNow(range))
        ) {
            return themes;
        }

        themes.push(id);

        return themes;
    }, []);
};

Stationery.pickerOrder = function (isCorp, lang, options = {}) {
    var returnArr = [], themeName, themeObj,
        i, themesInOrder = Stationery.templates.supportedThemesInOrder;

    if (options.enableThemeDateRange) {
        return this.getThemesForToday(isCorp, lang);
    }

    for (i in themesInOrder) {
        themeName = themesInOrder[i];
        themeObj = Stationery.stationeryThemeConfig.stationeryThemes[themeName];

        if (!themeObj || (!isCorp && themeObj.restricted) || !themeObj.supportedLang_regex.test(lang)) {
            continue;
        } else {
            returnArr.push(themeName);
        }
    }

    return returnArr;
};

Stationery.clearBehavior = function() {
    var behaviorCssNode, behaviorJSNode;

    Stationery.behavior && Stationery.behavior.deinit && Stationery.behavior.deinit();
    delete Stationery.behavior;

    behaviorCssNode = document.head.querySelector("#behavior_css");
    behaviorJSNode = document.head.querySelector("#behavior_js");

    behaviorCssNode && behaviorCssNode.remove();
    behaviorJSNode && behaviorJSNode.remove();
};

Stationery.applyBehavior = function(themeId) {
    this.clearBehavior();

    var themeObj = JSON.parse(JSON.stringify(this.stationeryThemeConfig.stationeryThemes[themeId])),
        stationeryBehaviorPath = this.stationeryThemeConfig.stationeryBehaviorPath, linkNode, scriptNode;

    if (themeObj.extra_css) {
        // apply css
        linkNode = document.createElement("link");
        linkNode.rel = "stylesheet";
        linkNode.id = "behavior_css";
        linkNode.href = stationeryBehaviorPath + themeObj.extra_css;
        linkNode.type = "text/css";
        document.head.appendChild(linkNode);
    }

    if (themeObj.extra_js) {
        // apply js
        scriptNode = document.createElement("script");
        scriptNode.id = "behavior_js";
        scriptNode.src = stationeryBehaviorPath + themeObj.extra_js;
        document.head.appendChild(scriptNode);
    }
};

Stationery.supportedBehavior = {
    shake: function() {
        Stationery.behavior && Stationery.behavior.shake && Stationery.behavior.shake();
    }
};

Stationery.deviceMetrics = {
    lang: "en-US"
};