var YMMessageRead = (function() {

    const _isAMPMessage = Boolean(YMConfig.ampBody);

    // [MI-23549] Hide message body until the content is settled down (dark mode conversion, resizing),
    // to reduce the chance of flashing and jumpiness on initial open or reload (e.g dark mode toggle).
    if (!YMConfig.isMessagePreviewMode && !_isAMPMessage) {
        document.documentElement.style.opacity = '0';
    }
    
    let _scaleFactor = 1, // ScaleFactor for Email content.
        _contentWrapper = document.getElementById("content-wrapper"),
        _attachmentsWrapper = document.getElementById("attachments"),
        _annotationButtonWrapper = document.createElement("div"),
        _content = document.getElementById("content"),
        _ymConfig = window.YMConfig,
        _msgPadding = _ymConfig.msgPadding,
        _skipMarkAsRead = false,
        _outerDivWidth100PercentEnabled = _ymConfig.outerDivWidth100PercentEnabled,
        _syncTimer,
        _zoomLevel = Math.round(_ymConfig.initialPercentTextSizeMultiplier * 100),
        // Email context-menu vars
        LONG_PRESS_DURATION = 600,
        _longPressTimer,

        _canAdjustWebViewHeight = false,
        _adjustWebViewHeightTimer,

        _resizeMessagePosted = false,

        _lastUpdatedAppWebViewHeight = -1,
        _longestPossibleLineBreakCharactersCount = 16,
        _viewMoreString = _ymConfig.viewMoreString,
        _viewLessString = _ymConfig.viewLessString,
        _hideQuotedText = _ymConfig.hideQuotedText,
        HIDDEN_ANNOTATION_CLASS = "annotation-hidden",
        _setOverflowingDivsWidth100OnOrientationChange = false,
        _bodyPaddingsAdded = false,
        _webViewOffsetWidth;

    const YIV_CONTENT_ID_SELECTOR = '#content > div[id]:first-child';

    const MAX_TEXT_CONTENT_LENGTH = 256;
    const AUTO_HEIGHT_OVERRIDE_CLASSNAME = "auto-height-override";
    const DEFAULT_ZOOM_LEVEL = 100;

    // [MI-24100] No-op hooks for form/link warnings to prevent runtime exceptions
    window.theMainWindow = {
        showFormWarning: () => {},
        showPasswordWarning: () => {},
        showLinkWarning: () => {},
    };

    const usingTextSizeAdjustPolyfill = () => typeof TextSizeAdjustPolyfill === "object";

    function skipMarkAsRead() {
        _skipMarkAsRead = true;
        return true;
    }

    function breakLongWordsInTable() {
        const acceptedFunction = node => {
            if (node.tagName === 'TABLE') {
                return NodeFilter.FILTER_ACCEPT;
            }
            return NodeFilter.FILTER_SKIP;
        };
        const walk = document.createTreeWalker(_content, NodeFilter.SHOW_ELEMENT, { acceptNode: acceptedFunction }, false);
        let tableNode = walk.nextNode();

        while (tableNode) {
            letLongWordsToBreakByAddingWBR(tableNode);
            tableNode = walk.nextSibling();
        }
    }

    function letLongWordsToBreakByAddingWBR(content) {
        const walk = document.createTreeWalker(content, NodeFilter.SHOW_TEXT, null, false);
        let n;
        while (n = walk.nextNode()) {
            giveOpportunityToBreakWordsIfWordsAreTooLong(n);
        }
    }

    function giveOpportunityToBreakWordsIfWordsAreTooLong(textNode) {
        const parentElementTagName = textNode.parentElement.tagName;
        if (parentElementTagName == "SCRIPT" || parentElementTagName == "STYLE") {
            return;
        }
        const lineBreakRegex = new RegExp("([\\S]{" + _longestPossibleLineBreakCharactersCount + "})", "g");
        const wordBreakIndices = [];
        let result;
        while (result = lineBreakRegex.exec(textNode.data)) {
            wordBreakIndices.push(lineBreakRegex.lastIndex);
        }
        if (wordBreakIndices.length > 0) {
            let lastIncludedStringIndex = textNode.data.length;
            for (let i = wordBreakIndices.length - 1; i >= 0; i--) {
                const newTextNodeTexts = textNode.data.substring(wordBreakIndices[i], lastIncludedStringIndex);
                textNode.after(document.createTextNode(newTextNodeTexts));
                textNode.after(document.createElement("wbr"));
                lastIncludedStringIndex = wordBreakIndices[i];
            }
            textNode.data = textNode.data.substring(0, lastIncludedStringIndex);
        }
    }

    /**
     * Add paddings to Emails that has any texts aligned too closely to the left or right. Otherwise, display edge-to-edge full screen Email viewing experience.
     */
    function addPaddingToAppropriateEmails() {
        const documentElementScrollWidth = document.documentElement.scrollWidth,
            documentElementOffsetWidth = document.documentElement.offsetWidth;
        // Go iterate through notes of the tree that has text nodes.
        let n;
        // Don't need to walk the attachments tray
        const walk = document.createTreeWalker(_content, NodeFilter.SHOW_TEXT, null, false);
        while (n = walk.nextNode()) {
            if (/\S/.test(n.data)) {
                if (isTextVisibleAndAlignedTooCloseToLeftOrRight(n.parentNode, documentElementScrollWidth)) {
                    document.body.style.padding = `${_msgPadding}px`; // Apply paddings in the document body so that it doesn't interfere with transform scale function. Also with paddings, scrollWidth doesn't include right padding when it is overflowing.
                    _webViewOffsetWidth = documentElementOffsetWidth - _msgPadding * 2;
                    _bodyPaddingsAdded = true;
                    break;
                }
            }
        }
        if (_attachmentsWrapper.querySelectorAll('.attachment').length > 0) {
            if (_bodyPaddingsAdded) { // Only need padding between attachment tray and the content because document.body's paddings will already cover left, right, bottom paddings of the attachment tray.
                _attachmentsWrapper.style.paddingTop = `${_msgPadding}px`;
            } else { // There is no padding around attachments tray. Add it all around it.
                _attachmentsWrapper.style.padding = `${_msgPadding}px`;
            }
        }
        _attachmentsWrapper.style.maxWidth = (documentElementOffsetWidth - _msgPadding * 2 - 8) + 'px'; // 100% viewport - msgPadding padding on each side - 8px pill margin.
    }

    function isTextVisibleAndAlignedTooCloseToLeftOrRight(element, rightHandBound) {
        const computedStyle = getComputedStyle(element);
        const MINIMUM_ALIGNED_TEXT_EDGE_SPACING = 5,
            MINIMUM_CONSIDERABLE_FONT_SIZE = 4,
            MINIMUM_CONSIDERABLE_LENGTH = 4;
        if (computedStyle.display == 'none' || computedStyle.visibility == 'hidden' || parseFloat(computedStyle.fontSize) < MINIMUM_CONSIDERABLE_FONT_SIZE) {
            return false;
        }
        // If text alignment is centered, ignore it. Many labels in html with a few words have left edge to right edge with centered text, hence don't apply paddings because of a centered text.
        if (computedStyle.textAlign == 'center' || computedStyle.textAlign == '-webkit-center') {
            return false;
        }
        const boundingRect = element.getBoundingClientRect();
        if (boundingRect == null) {
            return false;
        }
        // leftSpace and rightSpace are position of the text node inside contentWrapper. ContentWrapper can overflow because of the Email's content so use contentWrapper's scrollWidth get the right most point.
        // getBoundingClientRect.left or right includes margin, but not padding and borderWidth.
        const leftSpace = boundingRect.left + parseFloat(computedStyle.paddingLeft) + parseFloat(computedStyle.borderTopWidth),
            rightSpace = boundingRect.right - parseFloat(computedStyle.paddingRight) - parseFloat(computedStyle.borderTopWidth),
            height = boundingRect.bottom - boundingRect.top,
            width = boundingRect.right - boundingRect.left;
        if (parseFloat(width) < MINIMUM_CONSIDERABLE_LENGTH || parseFloat(height) < MINIMUM_CONSIDERABLE_LENGTH) {
            return false;
        }

        // Check if text is too close to left or right (right for right-to-left languages also).
        if (leftSpace <= MINIMUM_ALIGNED_TEXT_EDGE_SPACING || rightSpace >= rightHandBound - MINIMUM_ALIGNED_TEXT_EDGE_SPACING) {
            return true;
        }
        return false;
    }

    function getDocumentScrollWidthExcludingPaddings() {
        const offsetWidth = _webViewOffsetWidth;
        const scrollWidth = document.documentElement.scrollWidth;

        if (_bodyPaddingsAdded && scrollWidth > offsetWidth) {
            return scrollWidth - _msgPadding; // When scrollWidth > offsetWidth with left right paddings, we can guarentee that scrollWidth contains the left padding and not right padding. If content offsetWidth is 398; left & right paddings are 16px; and scrollWidth is 399, we know that the actual scrollWidth is 383 because 399 includes the 16px left padding.
        } else {
            return scrollWidth;
        }
    }

    /**
     * To scale the content to fit the width of the screen,
     * transform the content wrapper with the ratio of document.body.offsetWidth / document.body.scrollWidth.
     */
    const scaleContentSize = function() {

        if (!isDomContentReady()) return;

        let webContentWidth;
        let containerWidth = _webViewOffsetWidth;

        if (_contentWrapper.style.transform.length > 0) {
            _contentWrapper.style.width = `${containerWidth}px`; // Changing orientation may require rescaling and need this width set so that texts get wrapped to the next line at the correct place.
            var curTransform = new WebKitCSSMatrix(getComputedStyle(_contentWrapper).webkitTransform);
            webContentWidth = getDocumentScrollWidthExcludingPaddings();
            if (Math.abs(curTransform.a * webContentWidth - containerWidth) < 0.01) {
                // Current transform is good enough.
                if (webContentWidth <= containerWidth) {
                    _contentWrapper.style.width = `${containerWidth}px`;
                    return;
                } else {
                    _contentWrapper.style.width = `${_webViewOffsetWidth / _scaleFactor}px`;
                    return;
                }
            }
            // Reset the previous transform and re-scale it.
            _annotationButtonWrapper.style.transform = 'none';
            _attachmentsWrapper.style.transform = 'none';
            _contentWrapper.style.transform = 'none';
            _scaleFactor = 1.0;
            webContentWidth = getDocumentScrollWidthExcludingPaddings();
            containerWidth = _webViewOffsetWidth;
        } else {
            _contentWrapper.style.width = "";
            webContentWidth = getDocumentScrollWidthExcludingPaddings();
        }

        if (webContentWidth <= containerWidth) {
            _contentWrapper.style.width = `${containerWidth}px`;
            return;
        }
        _scaleFactor = containerWidth / webContentWidth;
        _contentWrapper.style.transform = `scale(${_scaleFactor})`;
        _contentWrapper.style.transformOrigin = '0 0';
        _contentWrapper.style.width = `${_webViewOffsetWidth / _scaleFactor}px`; // Need to set the width so that the container of mainly texts based Emails without fixed width would stretch all the way to the right.

        //[MI-20832] Handle the case where the content width constantly exceeds the container width. Can happen e.g if there is an element
        // that sets width to 100% and adds padding at the same time. We have just resized the contentWrapper to match the content width.
        // If the content width still exceeds the contentWrapper width, reset the container width.
        if (_contentWrapper.scrollWidth - _contentWrapper.offsetWidth > _msgPadding) {
            _contentWrapper.style.width = `${containerWidth}px`;
        }

        _annotationButtonWrapper.style.transform = `scale(${1 / _scaleFactor})`;
        _annotationButtonWrapper.style.transformOrigin = '0 0';

        _attachmentsWrapper.style.transform = `scale(${1 / _scaleFactor})`;
        _attachmentsWrapper.style.transformOrigin = '0 0';
    };


    function treeWalkerGoToNextSiblingNode(rootNode, n, walk) {
        n = walk.nextSibling(); // If there is no such node, this returns null and walk's current node is not changed.
        while (n == null) {
            n = walk.parentNode();
            if (n == rootNode) {
                return null;
            }
            n = walk.nextSibling();
        }
        return n;
    }

    /**
     * Set width of the elements that have no parent that has set width (anything under .style.width and .width) to be 100% so that mobile friendly Emails to be longer horizontally, making everything to be small when scaled down to fit width. Do this only when its offset Width is greater than content.clientWidth; Takes ~ 20milliseconds.
     */
    function setWidthOfOuterDivsToBe100Percent() {
        const walk = document.createTreeWalker(_content,
            NodeFilter.SHOW_ELEMENT,
            null,
            false);
        let n = walk.nextNode();
        while (n) {
            if (n.nodeName === "IMG") {
                n = treeWalkerGoToNextSiblingNode(_content, n, walk);
            } else {
                if (isThereValueForElementStyleWidthOrHeight(n.style.width) || isThereValueForElementWidthOrHeight(n.width)) {
                    if (n.offsetWidth > _webViewOffsetWidth) {
                        n.style.width = "100%";
                        n = treeWalkerGoToNextSiblingNode(_content, n, walk);
                    } else {
                        n = walk.nextNode();
                    }
                } else {
                    n = walk.nextNode();
                }
            }
        }
    }

    function isThereValueForElementStyleWidthOrHeight(styledWidthOrHeight) {
        if (styledWidthOrHeight == "") {
            return false;
        }
        if (styledWidthOrHeight.match(/^[0\.]+$|^[\d\.]+[a-z%]+$/i)) {
            return true;
        } else {
            return false;
        }
    }

    function isThereValueForElementWidthOrHeight(widthOrHeight) {
        if (widthOrHeight == "") {
            return false;
        }
        if (String(widthOrHeight).match(/^[\d\.]+([a-z%]+)?$/i)) {
            return true;
        } else {
            return false;
        }
    }

    function rescaleImagesOnOrientationChange() {
        const images = document.images;
        for (let i = 0; i < images.length; i++) {
            const image = images[i];
            const imageAspectRatio = image.dataset["overflowing-image-aspect-ratio"];
            if (image.complete && imageAspectRatio !== undefined) {
                image.style.height = image.clientWidth * imageAspectRatio + 'px';
            }
        }
    }

    /**
     * When changing orientation, images may or may not, start or continue to overflow.
     * We want to only apply our solution when image is overflowing - otherwise, it could have
     * unintended consequence. Revert it back to original values and see if we need to apply our solution again: maxWidth 100% and maintaining aspect ratio etc.
     * In MI-16412, aspect ratio of image of the car in portrait was maintained by external CSS rule height: auto which is not present in lanscape mode because this media style only applies for image width < 480px. Here, in lanscape, applying our solution from protrait when image isn't overflowing in lanscape squeezes the image.
     */
    function revertOverflowingImagesStyleAttributesToOriginalsOnOrientationChange() {
        const images = document.images;
        for (let i = 0; i < images.length; i++) {
            const image = images[i];
            if (image.dataset["overflowing-image-aspect-ratio"] !== undefined) {
                Object.assign(image.style, {
                    maxWidth: image.dataset["overflowing-image-original-style-max-width"],
                    height: image.dataset["overflowing-image-original-style-height"],
                    minHeight: image.dataset["overflowing-image-original-style-min-height"]
                });
                image.removeAttribute("data-overflowing-image-aspect-ratio");
                image.removeAttribute("data-overflowing-image-original-style-max-width");
                image.removeAttribute("data-overflowing-image-original-style-height");
                image.removeAttribute("data-overflowing-image-original-style-min-height");
            }
        }
    }

    /**
     * Set max-width of the images tables to be 100% so that mobile friendly images do not overflow.
     * Called from DomContentLoaded and Load.
     */
    function setWidthOfBigOverflowingImagesToBe100Percent() {
        const images = document.images;
        for (let i = 0; i < images.length; i++) {
            const image = images[i];
            if (image.dataset["overflowing-image-aspect-ratio"] === undefined && image.offsetWidth > _webViewOffsetWidth) {
                let shouldManuallyCalculateHeightOnLoad = false;
                let imageAspectRatio;
                if (isThereValueForElementStyleWidthOrHeight(image.style.height) || (image.getAttribute("height") && image.getAttribute("height") != "" && image.style.height != "auto" && image.getAttribute("height") != "auto")) {
                    shouldManuallyCalculateHeightOnLoad = true;
                    imageAspectRatio = image.clientHeight / image.clientWidth;

                    // Need to revert these to original in orientation change.
                    image.setAttribute("data-overflowing-image-original-style-max-width", image.style.maxWidth);
                    image.setAttribute("data-overflowing-image-original-style-height", image.style.height);
                    image.setAttribute("data-overflowing-image-original-style-min-height", image.style.minHeight);
                }
                Object.assign(image.style, {
                    maxWidth: '100%',
                    height: 'auto',
                    minHeight: 'auto'
                });
                //[MI-20784] If the image has a wrapper container that enforces min-height (intended for desktop browsers),
                // reset the wrapper container min-height constraint so that image can fully extend to its own height.
                if (image.parentElement && image.parentElement.children.length == 1) {
                    image.parentElement.style.minHeight = 'auto';
                }

                if (shouldManuallyCalculateHeightOnLoad) {
                    const rescaleImageFunction = function() {
                        image.setAttribute("data-overflowing-image-aspect-ratio", imageAspectRatio);
                        image.style.height = (image.clientWidth * imageAspectRatio) + 'px';
                    };
                    if (image.complete) {
                        rescaleImageFunction();
                    } else {
                        image.onload = function() {
                            // Josh filed a bug which cannot be reproduced but img with gif keep changing size. This seems only possible is image.onload gets called multiple times.
                            if (image.dataset["overflowing-image-aspect-ratio"] === undefined) {
                                rescaleImageFunction();
                            }
                        };
                        image.onerror = rescaleImageFunction;
                    }
                    continue;
                }
            }
            image.onload = function() {
                // If image with width, height = 1 is returned, that means that the image no longer exist in the server and placeholder image is returned. Prevent this image being a gigantic square by enforcing height to be 0. [MI-16562]
                if (image.naturalWidth == 1 && image.naturalHeight == 1) {
                    image.style.height = '0px';
                } else if (imageAspectRatioNeedsCorrection(image)) {
                     image.classList.add(AUTO_HEIGHT_OVERRIDE_CLASSNAME);
                }
            };

            if (image.complete && imageAspectRatioNeedsCorrection(image)) {
                image.classList.add(AUTO_HEIGHT_OVERRIDE_CLASSNAME);
            }
        }
    }

    /**
      * Checks if image DOM aspect ratio matches image intrinsic aspect ratio
      * @param image DOM element
      * @returns Boolean true if image aspect ratio needs adjustment
     */
    function imageAspectRatioNeedsCorrection(image) {
        const imageIsLoadedAndHasContent = image.complete && image.naturalWidth * image.naturalHeight > 1;
        if (!imageIsLoadedAndHasContent) return false;

        const {width: imageWidth, height:imageHeight} = elementContentSize(image);
        const imageIsVisible = imageWidth * imageHeight > 1;
        if (!imageIsVisible) return false;

        const intrinsicAspectRatio = image.naturalWidth / image.naturalHeight;
        const domAspectRatio = imageWidth / imageHeight;
        const aspectRatioOffByPercent = Math.abs(intrinsicAspectRatio - domAspectRatio) / intrinsicAspectRatio;

        // If aspect ratio is off by +5% and image height is constrained by min-height, should apply correction
        // Why min-height check? Because that is what is breaking aspect ratio for MI-17458, MI-15433, MI-16412, MI-21181,
        // when image width is constrained to viewport width (100% or less), but min-height is left as is. We make also
        // some additional checks in case image distortion is intentional, e.g. when stretching a small colored
        // image as a horizontal divider.
        const needsCorrection = aspectRatioOffByPercent > 0.05 &&
                               image.naturalWidth > imageWidth &&
                               imageHeight != parseFloat(getComputedStyle(image).maxHeight);
        return needsCorrection;
    }

    /**
      * Get size of the content box for a DOM element.
      * @param domElement
      * @returns object with width height
      */
    function elementContentSize(domElement) {
        let width = domElement.clientWidth; // clientWidth/Height exclude borders
        let height = domElement.clientHeight;

        // Subtract padding to get content box size
        const {vertical: verticalPadding, horizontal: horizontalPadding} = elementPadding(domElement);
        width = Math.max(0, width - horizontalPadding);
        height = Math.max(0, height - verticalPadding);

        return {width: width, height: height};
    }

    /**
      * Extracts padding values for a DOM element
      * @param domElement
      * @returns object with padding values as numbers
      */
    function elementPadding(domElement) {
        const toNumber = value => isNumber(value) ? value : 0;
        const computedStyle = getComputedStyle(domElement);

        const top = toNumber(parseFloat(computedStyle.paddingTop));
        const bottom = toNumber(parseFloat(computedStyle.paddingBottom));
        const left = toNumber(parseFloat(computedStyle.paddingLeft));
        const right= toNumber(parseFloat(computedStyle.paddingRight));

        return {top: top, bottom: bottom, left: left, right: right, vertical: top + bottom, horizontal: left + right};
    }

    function setZoom(zoom, updateLayout = true) {
        _zoomLevel = zoom;

        if (!isDomContentReady()) {
            return true;
        }

        if (usingTextSizeAdjustPolyfill()) { // iPad
            TextSizeAdjustPolyfill.zoomTextContent({rootNode: _contentWrapper, zoomLevel: zoom});
        } else { // iPhone
            _contentWrapper.style.webkitTextSizeAdjust = String(zoom) + "%";
            document.documentElement.style.setProperty('--text-size-adjust', String(zoom) + "%");
        }

        if (updateLayout) {
            scaleAndSync();
        }
        
        return true;
    }
    
    const addXhrInterceptionAttributes = html => {
        const parser = new DOMParser();
        const doc = parser.parseFromString(html, 'text/html');

        doc.documentElement.setAttribute('allow-xhr-interception', '');
        doc.documentElement.setAttribute('data-amp-autocomplete-opt-in', '');
        doc.documentElement.setAttribute('report-errors-to-viewer', '');

        return doc.documentElement.outerHTML;
    };
    
    function createAndAppendAmpIframe(){
        const iframeDivRef = document.querySelector('#content');
        const iframeRef = document.createElement('iframe');
        iframeRef.setAttribute('id','amp-iframe');
        iframeRef.setAttribute('sandbox', "allow-scripts allow-forms allow-popups allow-popups-to-escape-sandbox");
        iframeRef.setAttribute('allow', "autoplay 'none'; camera 'none'; encrypted-media 'none'; fullscreen 'none'; geolocation 'none'; gyroscope 'none'; magnetometer 'none'; microphone 'none'; midi 'none'; payment 'none'; speaker 'none'; sync-xhr 'none'; vr 'none'");
        iframeRef.setAttribute('scrolling', "no");
        iframeDivRef.appendChild(iframeRef);
        return iframeRef;
    };
    
    async function loadAmpIfNeeded() {
        if (YMConfig.ampBody) {
            const iframeRef = createAndAppendAmpIframe();
            const messagingToken = getAmpMessagingToken();
            if (iframeRef && messagingToken) {
                iframeRef.src = YMConfig.ampViewerURL + encodeURIComponent(messagingToken);
                iframeRef.onload = function() {
                    iframeRef.contentWindow.postMessage({amp: addXhrInterceptionAttributes(YMConfig.ampBody) },'*');
                };
                
                const messaging = await YAHOO.libs.AMP.Messaging.Messaging.waitForHandshakeFromDocument(window,iframeRef.contentWindow,'null',messagingToken);
                messaging.setDefaultHandler((name, data, rsvp) => {
                });
                messaging.registerHandler('xhr', NativeMessaging.xhrHandler);
                messaging.registerHandler('documentHeight', handleDocumentHeightRequest);
                messaging.sendRequest('visibilitychange', {state: "visible"}, true);
            }
        }
        return true;
    }
    
    const handleDocumentHeightRequest = (name, data) => {
        const iframeDivRef = document.querySelector('#content');
        if (data && data.height) {
            iframeDivRef.style.height = data.height + 'px';
            syncScrollHeightWithApp();
        }
    }

    const syncScrollHeightWithApp = () => {

            if (!isDomContentReady()) return;
            if (!isNumber(_scaleFactor)) {
                scaleContentSize();
            }

            // documentOffsetHeight and attachmentWrapperHeight are only two reliable heights we can use to calculate the actual height that the MessageDetailCell needs to display. Our _content element is transformed and the actual height of _content will be its true offsetHeight * _scaleFactor. _content's height can be anything due to any crazy html rules out there and may not add up with _attachmentsWrapper's height to be documentOffsetHeight (I have tested this statement, and it is true). We create attachmentWrapper and so attachmentWrapperHeight is not transformed and reliable. So use documentOffsetHeight attachmentWrapperHeight to calculate _content's height then add all back up to calculate the actual, final height.
            const messagePaddings = _bodyPaddingsAdded ? _msgPadding * 2.0 : 0;
            const documentOffsetHeightWithoutPaddings = document.documentElement.offsetHeight - messagePaddings;
            const attachmentWrapperHeight = _attachmentsWrapper.offsetHeight;
            const annotationButtonWrapperHeight = _annotationButtonWrapper.offsetHeight;

            let newWebViewHeight = (documentOffsetHeightWithoutPaddings - attachmentWrapperHeight) * _scaleFactor + attachmentWrapperHeight + annotationButtonWrapperHeight + messagePaddings;

            if (_lastUpdatedAppWebViewHeight != newWebViewHeight) {
                _lastUpdatedAppWebViewHeight = newWebViewHeight;
                _resizeMessagePosted = true;
                postMessage({
                    action: "resize",
                    height: newWebViewHeight
                });

                _canAdjustWebViewHeight = true;
                scheduleWebViewHeightCheck({delay: 500});

                // [MI-23549] Unhide message body - hidden until the content is settled down
                if (document.documentElement.style.opacity === '0') {
                    setTimeout(() => {
                        clearDocumentOpacity()
                    }, _bodyPaddingsAdded ? 100 : 50);
                }
            }
        },
        previewAttachment = function(attachmentEl) {
            var filename = attachmentEl.getAttribute("data-filename"),
                filesize = attachmentEl.getAttribute("data-filesize");

            window.location = "ios://preview-attachment?filename=" + encodeURIComponent(filename) + "&filesize=" + encodeURIComponent(filesize);
        },
        isEmailLink = function(el) {
            return el.nodeName.toLowerCase() === "a" && el.href && el.href.startsWith("mailto:");
        },
        isAnnotationButton = function(el) {
            return el.nodeName.toLowerCase() === "a" && el.getAttribute('data-action') === "toggle-annotation";
        },
        addEventListeners = function() {

            document.querySelector("#attachments").addEventListener("click", function(e) {
                const attachment = e.target.closest(".attachment");
                if (attachment) {
                    previewAttachment(attachment);
                }
            });

            _content.addEventListener("click", function(e) {
                const anchorLink = e.target.closest('a[href^="#"]');
                if (anchorLink) {
                    e.preventDefault();
                    handleAnchorLinkTap(anchorLink);
                    return;
                }

                const target = e.target,
                    inlineAttachment = target.closest("[data-ymail=inline-attachment]"),
                    anchor = target.closest("a");


                // inline attachment ([MI-23823] Anchors take precedence over attachments)
                if (inlineAttachment && !anchor) {
                    previewAttachment(inlineAttachment);
                    return;
                }

                // annotation
                if (isAnnotationButton(target)) {
                    toggleAnnotation(target);
                    return;
                }

                // Ignore links (opens draft if needed)
                if (!anchor) {
                    postMessage({
                        action: "contentTap"
                    });
                    return;
                }

                let textContent = anchor.textContent || '';

                // Check alt text in case an image was clicked
                for (let i = 0; i < anchor.childElementCount; i++) {
                    if (anchor.children[i].tagName === 'IMG' && anchor.children[i].alt) {
                        textContent = anchor.children[i].alt;
                        break;
                    }
                }

                textContent = textContent.trim().slice(0, MAX_TEXT_CONTENT_LENGTH);
                postMessage({
                    action: "linkTap",
                    textContent,
                    url: anchor.getAttribute("href")
                });
            });

            // Email context-menu on long-press
            document.body.addEventListener("touchstart", function(e) {
                let target = e.target;

                if (isEmailLink(target)) {
                    clearTimeout(_longPressTimer);
                    _longPressTimer = setTimeout(function() {
                        postMessage({
                            action: "emailMenu",
                            email: target.innerText
                        });
                    }, LONG_PRESS_DURATION);
                }
            });

            document.body.addEventListener('touchend', function() {
                clearTimeout(_longPressTimer);
            });

            window.addEventListener('scroll', function() {
                clearTimeout(_longPressTimer);
            });

            window.addEventListener("orientationchange", function() {
                setTimeout(() => checkForOrientationChange({orientationDidChange: true}), 0);
            });

            window.addEventListener("resize", function() {
                if (isMessageReadCleared()) return;
                const resizeMessagePosted = _resizeMessagePosted;
                _resizeMessagePosted = false;

                scheduleWebViewHeightCheck({delay: 50});

                // Update custom text zooming - original font size might have changed for some elements
                if (_zoomLevel != DEFAULT_ZOOM_LEVEL && usingTextSizeAdjustPolyfill()) {
                    if (Math.abs(_webViewOffsetWidth - getEffectiveWebViewOffsetWidth()) > 2) {
                        TextSizeAdjustPolyfill.updateZoomedText({debounce_ms: 100, completion: (updated) => {
                            updated && setTimeout(scaleAndSync, 100);
                        }});
                    }
                }

                const orientationChanged = checkForOrientationChange();
                if (!orientationChanged && !resizeMessagePosted) {
                    scaleAndSync();
                }

                // [MI-21858][iPad] top of message might get truncated on expanding to full-screen mode.
                // Appears to be a WebKit issue. Workaround - trigger complete page layout and re-render.
                forceDOMReflow();
            });
        },
        toggleAnnotation = function(target) {
            let targetDivs = document.querySelectorAll("." + target.getAttribute("data-target"));

            if (target.getAttribute("data-annotation-state") === "hidden") {
                targetDivs.forEach(function(div) {
                    div.classList.remove(HIDDEN_ANNOTATION_CLASS);
                });
                target.setAttribute("data-annotation-state", "shown");
                target.textContent = _viewLessString;
                postMessage({
                    action: "expandAnnotation"
                });
            } else {
                targetDivs.forEach(function(div) {
                    div.classList.add(HIDDEN_ANNOTATION_CLASS);
                });
                target.setAttribute("data-annotation-state", "hidden");
                target.textContent = _viewMoreString;
                postMessage({
                    action: "collapseAnnotation"
                });
            }
            setWidthOfBigOverflowingImagesToBe100Percent(); // Some images may have been hidden before finish loading images. For those images, in my testing, we need to rescale it again.
            scaleAndSync();
        },
        markAsRead = function() {
            if (_skipMarkAsRead) { return; }
            postMessage({
                action: "markAsRead"
            });
        },
        // From from Raven findImages(), not entirely sure what the "_origMsg_" checks are for...
        findImages = function() {
            var status = 1,
                images = document.images,
                links = document.links,
                lenImgs = images.length,
                lenLnks = links.length,
                rmvImgLen = 0,
                rmvLnkLen = 0,
                i;

            for (i = 0; i < lenImgs; i++) {
                let p = images[i];
                if (p.height == 1 && p.width == 1) {
                    rmvImgLen++;
                    continue;
                }
                while (p != document.documentElement) {
                    if (p.id == "_origMsg_") {
                        rmvImgLen++;
                        break;
                    }
                    p = p.parentNode;
                }
            }
            for (i = 0; i < lenLnks; i++) {
                let p = links[i];
                while (p != document.documentElement) {
                    if (p.id == "_origMsg_") {
                        rmvLnkLen++;
                        break;
                    }
                    p = p.parentNode;
                }
            }
            if (lenImgs - rmvImgLen > 0) {
                status |= 2;
            }
            if (lenLnks - rmvLnkLen > 0) {
                status |= 4;
            }
            return status;
        },
        hasImages = function() {
            if (!_ymConfig.isBlockImagesEnabled) {
                return false;
            }

            // Return a boolean
            return !!(findImages() & 2);
        },
        scaleAndSync = function() {
            if (!isDomContentReady()) return;
            setWidthOfBigOverflowingImagesToBe100Percent();
            scaleContentSize();
            syncScrollHeightWithApp();
        },
        // Sync height every second until page has finished loading (which is sometimes never)
        startSyncingHeight = function() {
            const synchBlock = () => {
                if (_webViewOffsetWidth != getEffectiveWebViewOffsetWidth()) {
                    checkForOrientationChange();
                } else {
                    scaleAndSync();
                }
            };
            // First re-sync after 50ms, after that every second until message load completes
            setTimeout(synchBlock, 50);
            _syncTimer = setInterval(synchBlock, 1000);
        },
        stopSyncingHeight = function() {
            clearTimeout(_syncTimer);
            setTimeout(syncScrollHeightWithApp, 500);
            
            return true;
        },
        // dynamically adds print header to webview
        addPrintHeader = function(htmlString) {
            var headerID = "print-header";
            if (document.getElementById(headerID)) {
                return false;
            }
            var headerElement = document.createElement("div");
            headerElement.id = headerID;
            headerElement.innerHTML = htmlString;
            _contentWrapper.insertBefore(headerElement, _contentWrapper.firstChild);
            
            return true;
        },
        clearDocumentOpacity = function() {
            document.documentElement.style.opacity = '';
            return true;
        },
        addAnnotationButton = function(quotedDiv, classId, initialStateHidden) {
            let buttonContainer = _annotationButtonWrapper,
                parentNode = quotedDiv.parentNode,
                nextSibling = quotedDiv.nextSibling,
                dataAnnotationState = initialStateHidden ? "hidden" : "shown",
                buttonText = initialStateHidden ? _viewMoreString : _viewLessString;

            buttonContainer.className = 'annotation-button-wrapper';
            buttonContainer.innerHTML = "<a class='annotation-button' data-action='toggle-annotation' data-annotation-state='" + dataAnnotationState + "' data-target='" + classId + "'>" + buttonText + "</a>";

            if (nextSibling) {
                parentNode.insertBefore(buttonContainer, nextSibling);
            } else {
                parentNode.appendChild(buttonContainer);
            }
        },
        hideQuotedAnnotations = function() {
            let classId = _ymConfig.htmlAnnotationsQuoted;

            if (!classId) {
                return;
            }

            let quotedDivs = document.querySelectorAll("." + classId),
                length = quotedDivs.length;

            if (length > 0) {
                // only need one annotation button per message, below the last quoted element
                addAnnotationButton(quotedDivs[length - 1], classId, _hideQuotedText);

                if (_hideQuotedText) {
                    quotedDivs.forEach(function(div) {
                        div.classList.add(HIDDEN_ANNOTATION_CLASS);
                    });
                }
            }
        },
        // Local drafts may have file:// prefix for img src
        loadLocalFilesIfNeeded = function() {
            let imgNodes = document.querySelectorAll("img.ymail-preserve-class[src^='file:']");
            imgNodes.forEach(function(node) { replaceAttributeValue(node, "src", "file:", "localfile:"); });
        },
        replaceAttributeValue = function(node, attribute, toReplace, replacement) {
            let attributeValue = node.getAttribute(attribute);
            node.setAttribute(attribute, attributeValue.replace(toReplace, replacement));
        },
        isDomContentReady = function() {
            // _webViewOffsetWidth gets initialized in DOMContentLoaded event handler
            return !isMessageReadCleared() && isNumber(_webViewOffsetWidth);
        },

        isMessageReadCleared = () => hasValue(window.YMMessageRead) === false,

        /* MI-24873: remove any data detector features from attachment filenames; timeout is needed so removal occurs sufficiently after data detectors are added by the os */
        removeAttachmentDataDetectors = () => setTimeout(() => document.querySelectorAll("#attachments-files a[x-apple-data-detectors]").forEach(a => a.removeAttribute("href")), 100),
    
        afterDomContentLoaded = function() {

            // [MI-24507] Safe message body load
            if (typeof _messageBody !== 'undefined' && hasValue(_content)) {
                _content.innerHTML = _messageBody;
            }

            if (document.documentElement.offsetWidth == 0) {
                GlobalExceptionHandler.log("DOMContentLoaded but document.documentElement.offsetWidth == 0, cannot render message", GlobalExceptionHandler.level.warning);
            }

            // [MI-23599] Splunk logs show _attachmentsWrapper is sometimes uninitialized
            // [MI-24507] The issue seems to be caused by the message _content body being not well-formed HTML
            _contentWrapper = document.getElementById("content-wrapper");
            if (validateElementPresenceRequestingSafeMessageReloadIfNeeded(_contentWrapper, "content-wrapper")) return;

            _attachmentsWrapper = document.getElementById("attachments");
            if (validateElementPresenceRequestingSafeMessageReloadIfNeeded(_attachmentsWrapper, "attachments")) return;

            _content = document.getElementById("content");
            if (validateElementPresenceRequestingSafeMessageReloadIfNeeded(_content, "content")) return;

            removeAttachmentDataDetectors();
            repairCssRulesIfNeeded();
            loadLocalFilesIfNeeded();

            // [MI-24297] Disable TD element 'nowrap' attribute if content is too wide (exceeds by more than 50% device width),
            // to make text wrap around and prevent tiny font sizes.
            if (_content.scrollWidth > innerWidth * 1.5 && hasValue(_content.querySelector('td[nowrap]'))) {
                _content.classList.add('disable-td-nowrap');
            }

            _webViewOffsetWidth = document.documentElement.offsetWidth;

            if (window.YMColorInverter && !_isAMPMessage) {
                YMColorInverter.messageReadDarkModeInvertColors(_ymConfig.darkModeBackgroundColorRGB);
            }

            breakLongWordsInTable();
            // Set fixed width so that long words with 'word-wrap: break-word' do not dictate the actual length of the Email.
            addPaddingToAppropriateEmails();
            if (_outerDivWidth100PercentEnabled == true) {
                setWidthOfOuterDivsToBe100Percent();
                if (Math.abs(window.orientation) === 90) { // Landscape orientation
                    _setOverflowingDivsWidth100OnOrientationChange = true;
                }
            }

            overrideFixedTableLayoutIfNeeded();

            // When we are using default 100% as webkitTextSizeAdjust, do not set it as 100%, because by default it is set to default and Apple automatically make some fonts larger with their own unique algoirthm.
            if (_zoomLevel != DEFAULT_ZOOM_LEVEL) {
                setZoom(_zoomLevel, /*updateLayout:*/false);
            }

            setWidthOfBigOverflowingImagesToBe100Percent();
            hideQuotedAnnotations();
            scaleAndSync();
            addEventListeners();
            markAsRead();
            startSyncingHeight();
            forceDOMReflow();

            // Uncomment the next line to see page state and timing in the XCode console
            //setTimeout(() => GlobalExceptionHandler.logPageStatus("page initialized"), 0);
        },
        afterWindowLoad = function() {
            // Load AMP message HTML into the iframe, after the webview completely loaded.
            loadAmpIfNeeded();
            scaleAndSync();
            stopSyncingHeight();
            setTimeout(function() {
                postMessage({
                    hasImages: hasImages(),
                    action: "windowLoaded"
                });
            }, 0);
        },
        postMessage = function(data) {
            window.webkit.messageHandlers.messageDetailScriptHandler.postMessage(data);
        },
        getAmpMessagingToken = () => {
            const bytes = new Uint8Array(32);
            window.crypto.getRandomValues(bytes);
            const bytesStr = String.fromCharCode.apply(null, Array.from(bytes));
            return btoa(bytesStr);
        };
    // Uncomment to enable logging to window._msgs array for debugging
    // window._logMsg = function(string) {
    //     if (!window._msgs) {
    //         window._msgs = [];
    //     }
    //     window._msgs.push(string);
    // };
    // When html and scripts are loaded
    document.addEventListener("DOMContentLoaded", function(event) {
        if (isMessageReadCleared()) return;
        // MI-16209: execute logic in next execution queue to avoid the bug where document.body.offsetWidth == 0
        setTimeout(afterDomContentLoaded, 0);
    });

    window.addEventListener("load", function() {
        if (isMessageReadCleared()) return;
        // MI-16209: execute logic in next execution queue to ensure it comes after afterDomContentLoaded
        setTimeout(afterWindowLoad, 0);
    });

    // [MI-24507] Check if document structure is valid, and try to recover using safe message body reload
    function validateElementPresenceRequestingSafeMessageReloadIfNeeded(element, elementId) {
        let requestingSafeMessageReload = false;

        if (hasValue(element)) return requestingSafeMessageReload;

        const messageAlreadyLoadedInSafeMode = typeof _messageBody !== 'undefined';
        if (messageAlreadyLoadedInSafeMode) {
            GlobalExceptionHandler.log(`Document structure invalid, no element with id '${elementId}'`, GlobalExceptionHandler.level.error);
            return requestingSafeMessageReload;
        }

        // Document structure is compromised, request safe message body reload
        requestingSafeMessageReload = true;
        postMessage({
            missingElement: elementId,
            action: "needsSafeMessageBodyInjection",
        });

        return requestingSafeMessageReload;
    }

    // There are some rare cases where documentElement.offsetWidth remains the same even after
    // "orientationchange" event has fired (observed on iPad with iOS14). The optional parameter
    // orientationDidChange can be used to signal that orientation change did happen, without
    // checking for documentElement.offsetWidth change.
    function checkForOrientationChange({orientationDidChange = false} = {}) {
        if (!isDomContentReady()) return false;
        const newWebViewOffsetWidth = getEffectiveWebViewOffsetWidth();

        // If document's width didn't change we did not change orientation so just return here.
        if (!orientationDidChange && newWebViewOffsetWidth == _webViewOffsetWidth) {
            return false;
        }

        // The actual resizing for different orientation.
        _webViewOffsetWidth = newWebViewOffsetWidth;

        if (_setOverflowingDivsWidth100OnOrientationChange === true) {
            setWidthOfOuterDivsToBe100Percent();
            _setOverflowingDivsWidth100OnOrientationChange = false;
        }
        revertOverflowingImagesStyleAttributesToOriginalsOnOrientationChange();
        setWidthOfBigOverflowingImagesToBe100Percent();

        scaleAndSync();
        setTimeout(orientationChanged, 500);

        return true;
    }

    // Images deep inside the Dom would not have their 100% width updated to the new value in orientation change. Also in a few Emails, document's height may be incorrect right after orientation change resulting in white bottom space. Width of the Email is not going to change here so we can safely only update the height. Below code is harmless.
    function orientationChanged() {
        rescaleImagesOnOrientationChange();
        syncScrollHeightWithApp();
    }

    function repairCssRulesIfNeeded() {
        repairCssRulesWithDuplicatedYIVContentIdSelectors();
        overrideCSSRulesRemovingContent();
    }

    // MI-20822 There is a long outstanding backend YIV CSS rewriting bug MAS-14687 that breaks message CSS when comments are present in CSS styles.
    // Instead of being completely removed, a comment is replaced with a #yivContentId selector, which breaks the immediately following rule.
    // E.g. a CSS snippet:
    //    @media only screen and (max-width:600px) {
    //      /* Footer mobile */
    //      .link_footer {line-height: 21px!important;}
    //    }
    // when rewritten by YIV might look like this:
    //    @media only screen and (max-width:600px) {
    //      #yiv5336030765
    //      #yiv5336030765 .yiv5336030765link_footer {line-height: 21px!important;}
    //    }
    // As a workaround, we detect these broken rules and repair selectors by replacing duplicated #yivContentId selectors with a single #yivContentId.
    // If we go back to the example, the broken selector rule
    // #yiv5336030765 #yiv5336030765 .yiv5336030765link_footer {line-height: 21px!important;}
    // becomes
    // #yiv5336030765 .yiv5336030765link_footer {line-height: 21px!important;}
    // as intended.
    //
    function repairCssRulesWithDuplicatedYIVContentIdSelectors() {
        const yivContentElement = document.querySelector(YIV_CONTENT_ID_SELECTOR);
        const yivContentId = yivContentElement && yivContentElement.id;

        if (!yivContentId) return;

        const doubleContentIdSelectorPrefix = `#${yivContentId} `.repeat(2);
        let adjustedRules = [];

        for (const styleSheet of document.styleSheets || []) {
            for (const rule of styleSheet.cssRules || styleSheet.rules || []) {
                if ((rule.cssText || '').includes(doubleContentIdSelectorPrefix)) {
                    const adjustedRule = rule.cssText.replace(new RegExp(`(#${yivContentId}\\s+){2,}`, 'g'), `#${yivContentId} `);
                    // [MI-21592] yivContentId should never be the only selector
                    if (new RegExp(`${yivContentId}\\s+\\{`).test(adjustedRule) == false) {
                        adjustedRules.push(adjustedRule);
                    }
                }
            }
        }

        if (adjustedRules.length == 0) return;

        let repairedStyle = document.createElement('style');
        repairedStyle.id = 'MAS-14687-repaired-style';
        repairedStyle.textContent = adjustedRules.join('\n');
        document.head.appendChild(repairedStyle);
    }

    // [MI-22054] Another backend YIV CSS rewriting bug where the main content container is removed,
    // (display: none), which should never be the case. We detect this condition, and override
    // violating CSS rules.
    function overrideCSSRulesRemovingContent() {
        const yivContentElement = document.querySelector(YIV_CONTENT_ID_SELECTOR);

        const needsOverride = yivContentElement && getComputedStyle(yivContentElement).display == 'none';
        if (!needsOverride) return;

        const yivContentId = yivContentElement && yivContentElement.id;
        if (!yivContentId) return;

        let overrideStyle = document.createElement('style');
        overrideStyle.id = 'MI-22054-yivContentId-overrides';
        overrideStyle.textContent = `#content #${yivContentId} {display: block !important;}`;
        document.head.appendChild(overrideStyle);
    }

    // document.documentElement.offsetWidth is not constant and might change during page load cycle
    // The initial value might be less than the final one, resulting in shrunk content layout unless
    // we periodically check and update the value.
    function getEffectiveWebViewOffsetWidth() {
        if (!document.documentElement) return undefined;
        let width = document.documentElement.offsetWidth;
        if (_bodyPaddingsAdded) {
            width -= _msgPadding * 2.0;
        }
        return width;
    }

    function scheduleWebViewHeightCheck({delay = 0} = {}) {
        if (!_ymConfig.webViewHeightCorrectionEnabled) return;
        if (_adjustWebViewHeightTimer) {
            clearTimeout(_adjustWebViewHeightTimer);
        }
        _adjustWebViewHeightTimer = setTimeout(adjustWebViewHeightIfNeeded, delay);
    }

    // [MI-20163][iPad][iOS14] In device portrait mode a message might have a large space block bellow
    // the content. This is because the WebView height is not matching the one set with the resize message
    // post (_lastUpdatedAppWebViewHeight value). This appears to be a WKWebKit issue. As a workaround,
    // we can detect this height mismatch and apply a correction.
    function adjustWebViewHeightIfNeeded() {
        if (!_ymConfig.webViewHeightCorrectionEnabled) return;
        if (_webViewOffsetWidth != getEffectiveWebViewOffsetWidth()) {
            checkForOrientationChange();
            return;
        }

        // Can try to adjust height only once after resize post message was sent
        if (!_canAdjustWebViewHeight || _lastUpdatedAppWebViewHeight < 1) return;
        _canAdjustWebViewHeight = false;

        // If actual viewport height exceeds the last updated height by more then 150px, apply a
        // correction by scaling the height by the off ratio.
        const viewportHeight = document.documentElement.clientHeight;
        const MAX_HEIGHT_DIFFERENCE = 150, SAFETY_SPACING = 64;
        if (viewportHeight - _lastUpdatedAppWebViewHeight > MAX_HEIGHT_DIFFERENCE) {
            const viewportHeightRatio = _lastUpdatedAppWebViewHeight / viewportHeight;
            const adjustedViewPortHeight = Math.ceil(_lastUpdatedAppWebViewHeight * viewportHeightRatio) + SAFETY_SPACING;
            _resizeMessagePosted = true;
            postMessage({
                action: "resize",
                height: adjustedViewPortHeight
            });
        }
    }

    // A workaround for WKWebKit render bugs, like MI-21801, that can be fixed with a forced DOM reflow (recompute DOM layout).
    // Forced DOM reflow is sync and unnoticeable, yet it fixes a whole class of potential WKWebKit render issues.
    function forceDOMReflow() {
        _content.style.display = 'none';
        _content.getBoundingClientRect(); // flushes CSS style changes to the DOM render tree
        _content.style.display = '';
    }

    function measureMaxTextExtent() {
        let textNode, textMaxX = 0;

        const treeWalker = document.createTreeWalker(_content, NodeFilter.SHOW_TEXT, null, false);
        while (textNode = treeWalker.nextNode()) {
            if (!/\w/.test(textNode.nodeValue)) continue;
            const boundingRect = textNode.parentNode.getBoundingClientRect();
            const maxX = Math.ceil(boundingRect.x + boundingRect.width);
            textMaxX = Math.max(textMaxX, maxX);
        }

        return Math.ceil(textMaxX);
    }

    // [MI-22227][MI-19095] Some messages, CSS styled exclusively for desktop screen sizes, constrain
    // their content width and use overflow:hidden, while having text that extends the constrained
    // bounds. We can detect if text is truncated, and adjust the content width to avoid truncation.
    function adjustContentWidthIfNeeded() {
        const DEFAULT_WEBPAGE_WIDTH = 980;

        _content.style.removeProperty('min-width');
        if (_content.scrollWidth > DEFAULT_WEBPAGE_WIDTH) return;

        const savedContentWrapperWidth = _contentWrapper.style.width || '';
        _contentWrapper.style.width = 'auto';

        const contentWidth = Math.ceil(_content.scrollWidth);
        const maxContentWidth = Math.ceil(measureMaxTextExtent() / _scaleFactor);
        const truncatedContentWidth = maxContentWidth - contentWidth;

        _contentWrapper.style.width = savedContentWrapperWidth;

        if (truncatedContentWidth > _msgPadding + 1) {
            _content.style.minWidth = `${maxContentWidth}px`;
        }
    }

    // [MI-23121] Support anchor links - scroll to section
    function handleAnchorLinkTap(anchorLink) {
        const linkId = (anchorLink.getAttribute("href") || "").substring(1);
        if (linkId.length == 0) return;

        // YIV rewrites message node IDs by prefixing them with the YIV container id
        const yivContentElement = document.querySelector(YIV_CONTENT_ID_SELECTOR);
        const yivContentId = yivContentElement && yivContentElement.id;
        if (!yivContentId) return;

        //[https://www.w3.org/TR/html401/struct/links.html] "Destination anchors in HTML documents may be specified either
        // by the A element (naming it with the name attribute), or by any other element (naming with the id attribute)"
        const destinationSelector = `[id="${linkId}"],[id="${yivContentId}${linkId}"],a[name="${yivContentId}${linkId}"],a[name="${linkId}"]`;
        const destinationElement = document.querySelector(destinationSelector);

        if (destinationElement) {
            const ANCHOR_TOP_MARGIN = 16;
            const scrollY = Math.round(destinationElement.getBoundingClientRect().y) - ANCHOR_TOP_MARGIN;
            postMessage( {
                action: "scrollTo",
                scrollTop: scrollY
            });
        }
    }

    // [MI-23110] Fixed table layout indicates that all table columns have in advance known width, and that a table does
    // need to adjust to the actual cell content size. When fixed layout is used for a table that actually has dynamic
    // content width, it can result in a table content overflowing, increasing the chance of having a broken layout.
    // We detect this state and change tableLayout from fixed to auto, if needed.
    function overrideFixedTableLayoutIfNeeded() {
        const fixedLayoutTables = Array.from(document.querySelectorAll('table'))
                                 .filter(table => getComputedStyle(table).tableLayout == 'fixed');

        for (const table of fixedLayoutTables) {
            const overflowWidth = table.scrollWidth - table.offsetWidth;
            if (overflowWidth > _msgPadding) {
                GlobalExceptionHandler.log(`Override tableLayout fixed, make it auto as content overflows ${overflowWidth}px`, GlobalExceptionHandler.level.info);
                table.style.tableLayout = 'auto';
            }
        }
    }

    function isNumber(value) {
       return typeof value === 'number' && isFinite(value);
    }

    function hasValue(value) {
        return !(value === null || value === undefined);
    }

    // MARK: - Public methods
    // All methods below must return a value to avoid `Fatal error: Bug in WebKit: Received neither result or failure` when evaluated via `WKWebView.evaluateJavaScript()`
    return {
        skipMarkAsRead: skipMarkAsRead,
        setZoom: setZoom,
        addPrintHeader: addPrintHeader,
        clearDocumentOpacity: clearDocumentOpacity
    };
}());
