var YRSEComposeController = (function() {
    /* CLASS NAMES */

    var HEADER_CONTAINER_ID = "yahoo__header",
        TO_CONTAINER_ID = "yahoo__to",
        CC_CONTAINER_ID = "yahoo__cc",
        BCC_CONTAINER_ID = "yahoo__bcc",
        FROM_CONTAINER_ID = "yahoo__from",
        SUBJECT_CONTAINER_ID = "yahoo__subject",
        COMPOSE_AREA_ID = "yahoo__compose_area",
        EXPANDABLE_CONTAINER_ID = "yahoo__expandable",
        ATTACHMENT_AREA_ID = "yahoo__attachments",
        SIGNATURE_ID = "yahoo__signature",
        CONTACT_SELECT_TOOLTIP_ID = "yahoo__contact_select_description",
        CONTACT_DESELECT_TOOLTIP_ID = "yahoo__contact_deselect_description",
        ARIA_DESCRIBEDBY_ATTRIBUTE = "aria-describedby",
        ARIA_ROLE_ATTRIBUTE = "role",
        DATA_GUID_ATTRIBUTE = "data-guid",
        DATA_PATH_ATTRIBUTE = "data-path",
        DATA_NAME_ATTRIBUTE = "data-name",
        DATA_EMAIL_ATTRIBUTE = "data-email",
        DATA_CONTENT_ATTRIBUTE = "data-content",
        DATA_CONTENT_VALUE_NAME = "name",
        DATA_CONTENT_VALUE_EMAIL = "email",
        ATTACHMENT_DIV_CLASS_NAME = "attachment",
        FILE_ATTACHMENT_CLASS_NAME = "file-attachment",
        OUTLINE_ATTACHMENT_CLASS_NAME = "outline-attachment",
        ATTACHMENT_OVERLAY_CLASS_NAME = "attachment-overlay",
        ATTACHMENT_MORE_BUTTON_CLASS_NAME = "attachment-more",
        RECIPIENT_HEADER_CLASS_NAME = "recipient-header",
        RECIPIENT_LIST_CLASS_NAME = "recipient-list",
        RECIPIENT_CLASS_NAME = "recipient",
        INVALID_RECIPIENT_CLASS_NAME = "invalid",
        SELECTED_CLASS_NAME = "selected",
        EXPANDED_CONTAINER_CLASS_NAME = "expanded",
        ICON_EXPANDED_CLASS_NAME = "icon-expanded",
        FROM_ELEMENT_CLASS_NAME = "from-email",
        SUBJECT_INPUT_SELECTOR = "#" + SUBJECT_CONTAINER_ID + " input",
        SCROLL_DEBOUNCE_TIME = 100,

        /* PRIVATE PROPERTIES */
        _headerContainer,
        _toHeaderContainer,
        _expandableContainer,
        _fromHeaderContainer,
        _subjectFieldInput,
        _composeArea,
        _signatureContainer,
        _attachmentArea,
        _focusedInput,
        _dblClickRecipientTimer,
        _scrollIfNeededTimeout,
        _recipientCurrentId = 0,
        _recipientDraggedId,

        /* PRIVATE METHODS */

        _initElements = function() {
            _headerContainer = document.getElementById(HEADER_CONTAINER_ID);
            _toHeaderContainer = document.getElementById(TO_CONTAINER_ID);
            _expandableContainer = document.getElementById(EXPANDABLE_CONTAINER_ID);
            _fromHeaderContainer = document.getElementById(FROM_CONTAINER_ID);
            _subjectFieldInput = document.querySelector(SUBJECT_INPUT_SELECTOR);
            _composeArea = document.getElementById(COMPOSE_AREA_ID);
            _signatureContainer = document.getElementById(SIGNATURE_ID);
            _attachmentArea = document.getElementById(ATTACHMENT_AREA_ID);
        },

        // Signature container may be deleted and we still have a reference to it,
        // so must check if it also has a parent
        _getSignatureContainer = function() {
            if (_signatureContainer && _signatureContainer.parentNode) {
                return _signatureContainer;
            }

            _signatureContainer = null;
            return null;
        },

        _attachmentWithGuid = function(messageSection, guid) {
            return messageSection.querySelector("[" + DATA_GUID_ATTRIBUTE + "='" + guid + "']");
        },

        _getRecipients = function(headerId) {
            var recipientEls = _headerContainer.querySelectorAll("#" + headerId + " ." + RECIPIENT_CLASS_NAME),
                recipients = [];

            recipientEls.forEach(function(el) {
                recipients.push({
                    name: el.getAttribute(DATA_NAME_ATTRIBUTE),
                    email: el.getAttribute(DATA_EMAIL_ATTRIBUTE)
                });
            });

            return recipients;
        },

        _getSenderEmail = function() {
            if (_fromHeaderContainer) {
                return _fromHeaderContainer.querySelector("." + FROM_ELEMENT_CLASS_NAME + " span").textContent;
            }

            return "";
        },

        _getTranslatedBody = function() {
            var clonedBody = _composeArea.cloneNode(true),
                inlineImages = clonedBody.querySelectorAll("[" + DATA_GUID_ATTRIBUTE + "]"),
                // Overlays and Buttons
                extraElements = clonedBody.querySelectorAll("." + ATTACHMENT_OVERLAY_CLASS_NAME + ",." + ATTACHMENT_MORE_BUTTON_CLASS_NAME),
                signature = clonedBody.querySelector("#" + SIGNATURE_ID);

            inlineImages.forEach(function(el) {
                el.querySelector("img").src = "cid://" + el.getAttribute(DATA_GUID_ATTRIBUTE);
            });
            extraElements.forEach(function(el) {
                el.remove();
            });

            if (signature) {
                signature.removeAttribute("id");
            }

            return clonedBody.innerHTML;
        },

        _getGuids = function(messageSection) {
            var attachments = messageSection.querySelectorAll("[" + DATA_GUID_ATTRIBUTE + "]"),
                guids = [];
            attachments.forEach(function(el) {
                guids.push(el.getAttribute(DATA_GUID_ATTRIBUTE));
            });
            return guids;
        },

        _scrollIfNeeded = function() {
            if (_scrollIfNeededTimeout) {
                clearTimeout(_scrollIfNeededTimeout);
            }

            _scrollIfNeededTimeout = setTimeout(function() {
                var yOffset = YRSEInputHandler.getCursorScrollPositionIfNeeded();

                // Scrolling natively (el.scrollTo()) isn't 100% reliable, so have the app do it
                if (yOffset) {
                    _postNativeMsg("window:scroll", {
                        yOffset: yOffset
                    });
                }
            }, SCROLL_DEBOUNCE_TIME);
        },

        _linkify = function() {
            YRSELinkify.linkify(_composeArea);
        },

        _postNativeMsg = function(msg, data) {
            if (!window.webkit) {
                return;
            }

            window.webkit.messageHandlers.nativeMessageHandler.postMessage(JSON.stringify({
                message: msg,
                data: data || null
            }));
        },
    
        _createRecipientElement = function(name, email) {
            var recipientElement = document.createElement("li");
            recipientElement.className = RECIPIENT_CLASS_NAME;
            recipientElement.textContent = name + ",";
            recipientElement.setAttribute(DATA_NAME_ATTRIBUTE, name);
            recipientElement.setAttribute(DATA_EMAIL_ATTRIBUTE, email);
            recipientElement.setAttribute(DATA_CONTENT_ATTRIBUTE, DATA_CONTENT_VALUE_NAME);
            recipientElement.setAttribute(ARIA_ROLE_ATTRIBUTE, "option");   // to support aria-describedby announcements when VoiceOver is enabled
            recipientElement.setAttribute(ARIA_DESCRIBEDBY_ATTRIBUTE, CONTACT_SELECT_TOOLTIP_ID);
            recipientElement.id = "recipient" + _recipientCurrentId++;
            
            // Drag and Drop
            recipientElement.setAttribute("draggable", true);
            recipientElement.addEventListener("dragstart", _onRecipientDragStart);
            recipientElement.addEventListener("dragenter", _onRecipientDragEnter);
            recipientElement.addEventListener("dragover", _onRecipientDragOver);

            if (!YRSEEmailAddressValidator.isValidEmail(email)) {
                recipientElement.classList.add(INVALID_RECIPIENT_CLASS_NAME);
            }

            return recipientElement;
        },

        _createLineBreak = function() {
            var div = document.createElement("div");
            div.appendChild(document.createElement("br"));
            return div;
        },

        _isLineBreak = function(element) {
            return (element.nodeName.toLowerCase() === "div" && element.innerHTML === "<br>");
        },

        _createAttachmentContainer = function(path, guid, filename) {
            var attachmentContainer = document.createElement("div");
            attachmentContainer.setAttribute(DATA_GUID_ATTRIBUTE, guid);
            attachmentContainer.setAttribute(DATA_PATH_ATTRIBUTE, path);
            attachmentContainer.setAttribute(DATA_NAME_ATTRIBUTE, filename);
            attachmentContainer.className = ATTACHMENT_DIV_CLASS_NAME;
            return attachmentContainer;
        },

        _createMoreButton = function() {
            var moreButton = document.createElement("button");
            moreButton.className = ATTACHMENT_MORE_BUTTON_CLASS_NAME;
            return moreButton;
        },

        _validateEmail = function() {
            var numRecipients = _headerContainer.querySelectorAll("." + RECIPIENT_HEADER_CLASS_NAME + " ." + RECIPIENT_CLASS_NAME).length,
                numInvalidRecipients = _headerContainer.querySelectorAll("." + RECIPIENT_HEADER_CLASS_NAME + " ." + INVALID_RECIPIENT_CLASS_NAME).length;

            if (numRecipients !== numInvalidRecipients) {
                _postNativeMsg("email:valid");
            } else {
                _postNativeMsg("email:invalid");
            }
        },

        _updateAutoSuggest = function(inputEl) {
            // Encode payload as JSON
            var headerEl = inputEl.closest("." + RECIPIENT_HEADER_CLASS_NAME),
                recipientEls = inputEl.closest("#" + HEADER_CONTAINER_ID).querySelectorAll("." + RECIPIENT_CLASS_NAME),
                endpoints = [],
                msgData = {};

            recipientEls.forEach(function(el) {
                endpoints.push(el.getAttribute(DATA_EMAIL_ATTRIBUTE));
            });

            msgData = {
                query: inputEl.value,
                endpoints: endpoints,
                yOffset: headerEl.offsetTop + headerEl.clientHeight - document.body.scrollTop + 1
            };

            _focusedInput = inputEl;

            _postNativeMsg("contact:input", msgData);
        },

        _handleRecipientEnter = function(inputEl, updateFocusOnEmpty) {
            var value = inputEl.value.trim(),
                recipientHeader = inputEl.closest("." + RECIPIENT_HEADER_CLASS_NAME),
                focusInputSelector,
                recipientList,
                isHeaderExpanded;

            // Validate
            if (/\S/.test(value)) {
                recipientList = recipientHeader.querySelector("." + RECIPIENT_LIST_CLASS_NAME);
                recipientList.insertBefore(_createRecipientElement(value, value), recipientList.lastElementChild);
                inputEl.value = "";
                _postNativeMsg("contact:close");
                _postNativeMsg("event:log", {
                    name: "recipient-added",
                    fieldId: recipientHeader.id
                });
                _validateEmail();
            } else if (updateFocusOnEmpty) {
                // Jump to next visible tabbable element
                isHeaderExpanded = _headerContainer.classList.contains(EXPANDED_CONTAINER_CLASS_NAME);
                switch (recipientHeader.id) {
                    case TO_CONTAINER_ID:
                        focusInputSelector = isHeaderExpanded ? "#" + CC_CONTAINER_ID + " input" : SUBJECT_INPUT_SELECTOR;
                        break;
                    case CC_CONTAINER_ID:
                        focusInputSelector = isHeaderExpanded ? "#" + BCC_CONTAINER_ID + " input" : SUBJECT_INPUT_SELECTOR;
                        break;
                    default:
                        focusInputSelector = SUBJECT_INPUT_SELECTOR;
                        break;
                }
                _focusInput(_headerContainer.querySelector(focusInputSelector));
            }
        },

        _handleRecipientBackspace = function(inputEl) {
            var recipientHeader = inputEl.closest("." + RECIPIENT_HEADER_CLASS_NAME),
                selectedRecipient = recipientHeader.querySelector("." + RECIPIENT_LIST_CLASS_NAME + " ." + SELECTED_CLASS_NAME),
                recipientEls,
                lastRecipient;

            if (inputEl.value === "") {
                if (selectedRecipient) {
                    selectedRecipient.remove();

                    _postNativeMsg("event:log", {
                        name: "recipient-removed",
                        fieldId: recipientHeader.id
                    });

                    _updateAutoSuggest(inputEl);
                } else {
                    recipientEls = inputEl.closest("." + RECIPIENT_LIST_CLASS_NAME).querySelectorAll("." + RECIPIENT_CLASS_NAME);
                    if (recipientEls.length > 0) {
                        lastRecipient = recipientEls[recipientEls.length - 1];
                        _selectRecipientEl(lastRecipient);
                    }
                }
            }

            _validateEmail();
        },

        _clearSelected = function() {
            var selectedRecipients = _headerContainer.querySelectorAll("." + RECIPIENT_HEADER_CLASS_NAME + " ." + SELECTED_CLASS_NAME);

            selectedRecipients.forEach(function(el) {
                el.classList.remove(SELECTED_CLASS_NAME);
                el.setAttribute(ARIA_DESCRIBEDBY_ATTRIBUTE, CONTACT_SELECT_TOOLTIP_ID);
            });
        },
    
        _selectRecipientEl = function(el) {
            el.classList.add(SELECTED_CLASS_NAME);
            el.setAttribute(ARIA_DESCRIBEDBY_ATTRIBUTE, CONTACT_DESELECT_TOOLTIP_ID);
        },

        _focusInput = function(inputEl) {
            inputEl.style.display = "inline-block";
            inputEl.focus();
        },

        _onRecipientKeyup = function(e) {
            var inputEl = e.target,
                key = e.key || e.keyIdentifier;

            if (key === "Enter") {
                _clearSelected();
                _handleRecipientEnter(inputEl, true);
            } else if (key === "Backspace" || key === "U+0008") {
                _handleRecipientBackspace(inputEl);
            } else {
                _clearSelected();
            }
        },

        _addTrailingCommaToEmail = function(email) {
            if (email[email.length - 1] !== ",") {
                email += ",";
            }
            return email;
        },

        _removeTrailingCommaFromEmail = function(email) {
            if (email[email.length - 1] === ",") {
                email = email.substring(0, email.length - 1);
            }
            return email;
        },

        _onRecipientFocus = function(e) {
            var target = e.target,
                recipientEls = target.closest("." + RECIPIENT_LIST_CLASS_NAME).querySelectorAll("." + RECIPIENT_CLASS_NAME),
                lastRecipient,
                email;

            if (recipientEls.length > 0) {
                lastRecipient = recipientEls[recipientEls.length - 1];
                email = _addTrailingCommaToEmail(lastRecipient.textContent);
                lastRecipient.textContent = email;
            }
        },

        _onRecipientBlur = function(e) {
            var target = e.target,
                recipientEls = target.closest("." + RECIPIENT_LIST_CLASS_NAME).querySelectorAll("." + RECIPIENT_CLASS_NAME),
                lastRecipient,
                email;

            _clearSelected();
            _handleRecipientEnter(target);

            if (recipientEls.length > 0) {
                lastRecipient = recipientEls[recipientEls.length - 1];
                email = _removeTrailingCommaFromEmail(lastRecipient.textContent);
                lastRecipient.textContent = email;
                target.style.display = "none";
            }

            _postNativeMsg("contact:close");
        },

        _onRecipientInput = function(e) {
            var inputEl = e.target;
            _updateAutoSuggest(inputEl);
        },

        _onMessageInput = function(e) {
            // MI-16714: prevent clearing of text when using ios toolbar format options
            if (e.inputType.startsWith("format")) {
                return;
            }

            _scrollIfNeeded();
        },

        _onMessageKeyup = function(e) {
            // Prevent clearing of text when cmd+a shortcut is used
            if (e.key === "Meta") {
                return;
            }
            
            var key = e.key || e.keyIdentifier;

            if (key === "Enter" || key === " " || key === "U+0020") {
                _linkify();
            }

            _scrollIfNeeded();
        },

        _onSubjectFocus = function() {
            _clearSelected();
        },

        _onSubjectKeyup = function(e) {
            var key = e.key || e.keyIdentifier;

            if (key === "Enter") {
                _composeArea.focus();
            }
        },

        _onMessageFocus = function() {
            _clearSelected();
        },

        _pastePasteboard = function() {
            _postNativeMsg("paste:pasteboard");
        },

        _pasteImage = function(item) {
            var blob = item.getAsFile(),
                reader = new FileReader();

            reader.onload = function(event) {
                _postNativeMsg("paste:image", {
                    type: item.type,
                    dataURI: event.target.result
                });
            };

            reader.readAsDataURL(blob);
        },

        _onMessagePaste = function(event) {
            var items = event.clipboardData.items,
                types = event.clipboardData.types,
                length = 0,
                hasImgData,
                index,
                item;

            if (items) {
                length = items.length;
            } else if (types) {
                _pastePasteboard();
                event.preventDefault();
                return;
            }

            for (index = 0; index < length; index++) {
                item = items[index];
                if (item.kind === 'file' && item.type.match(/^image\//)) {
                    hasImgData = true;
                    _pasteImage(item);
                }
            }

            if (hasImgData) {
                event.preventDefault();
            }

            setTimeout(function() {
                _linkify();
            }, 0);
        },

        _onMessageBlur = function() {
            _linkify();
        },

        _onRecipientTouchStart = function(e) {
            if (e.target.classList.contains(RECIPIENT_CLASS_NAME)) {
                // Double click was performed
                if (_dblClickRecipientTimer) {
                    _onRecipientDblClick(e);
                } else {
                    // Execute as if single click if timeout expires
                    _dblClickRecipientTimer = (function(cachedEvent) {
                        return setTimeout(function() {
                            clearTimeout(_dblClickRecipientTimer);
                            _dblClickRecipientTimer = null;
                            _onRecipientSnglClick(cachedEvent);
                        }, 500);
                    })(e);
                }
            } else {
                _onRecipientSnglClick(e);
            }
        },

        _onRecipientSnglClick = function(e) {
            var target = e.target,
                inputEl = target.closest("." + RECIPIENT_HEADER_CLASS_NAME).querySelector("input");

            if (target.classList.contains(RECIPIENT_CLASS_NAME)) {
                if (target.classList.contains(SELECTED_CLASS_NAME)) {
                    _clearSelected();
                } else {
                    _clearSelected();
                    _selectRecipientEl(target);
                }
            }

            inputEl.style.display = "inline-block";
            inputEl.focus();
        },

        _onRecipientDblClick = function(e) {
            var target = e.target;

            clearTimeout(_dblClickRecipientTimer);
            _dblClickRecipientTimer = null;
            if (target.getAttribute(DATA_CONTENT_ATTRIBUTE) === DATA_CONTENT_VALUE_NAME) {
                target.textContent = target.getAttribute(DATA_EMAIL_ATTRIBUTE) + ",";
                target.setAttribute(DATA_CONTENT_ATTRIBUTE, DATA_CONTENT_VALUE_EMAIL);
            } else {
                target.textContent = target.getAttribute(DATA_NAME_ATTRIBUTE) + ",";
                target.setAttribute(DATA_CONTENT_ATTRIBUTE, DATA_CONTENT_VALUE_NAME);
            }
        },

        _onToLabelClick = function() {
            var transitionHeight;

            if (_headerContainer.classList.contains(EXPANDED_CONTAINER_CLASS_NAME)) {
                transitionHeight = _expandableContainer.offsetHeight;
                _expandableContainer.style.height = transitionHeight + "px";
                _headerContainer.classList.remove(ICON_EXPANDED_CLASS_NAME);
                setTimeout(function() {
                    _expandableContainer.style.height = "0px";
                }, 0);
            } else {
                _expandableContainer.style.display = "block";
                _expandableContainer.style.width = _toHeaderContainer.offsetWidth;
                transitionHeight = _expandableContainer.offsetHeight;
                _expandableContainer.style.height = "1px";
                _headerContainer.classList.add(EXPANDED_CONTAINER_CLASS_NAME);
                _headerContainer.classList.add(ICON_EXPANDED_CLASS_NAME);
                setTimeout(function() {
                    _expandableContainer.style.height = transitionHeight + "px";
                }, 0);
            }
        },

        _onDocumentTouchEnd = function(e) {
            var target = e.target;

            if (target.classList.contains(ATTACHMENT_MORE_BUTTON_CLASS_NAME)) {
                _showMoreMenu(target);
                e.preventDefault();
            } else if (target.classList.contains(ATTACHMENT_OVERLAY_CLASS_NAME)) {
                e.preventDefault();
            }
        },

        _onFromClick = function() {
            var fromElementRect = _fromHeaderContainer.querySelector("." + FROM_ELEMENT_CLASS_NAME).getBoundingClientRect();

            _postNativeMsg("menu:sender", {
                x: fromElementRect.left,
                y: fromElementRect.top,
                width: fromElementRect.width,
                height: fromElementRect.height
            });
        },

        _showMoreMenu = function(moreButton) {
            var attachmentElement = moreButton.closest("[" + DATA_GUID_ATTRIBUTE + "]"),
                buttonRect = moreButton.getBoundingClientRect();

            _postNativeMsg("menu:attachment", {
                guid: attachmentElement.getAttribute(DATA_GUID_ATTRIBUTE),
                disposition: moreButton.closest("#" + ATTACHMENT_AREA_ID) ? "outline" : "inline",
                type: attachmentElement.classList.contains(FILE_ATTACHMENT_CLASS_NAME) ? "file" : "image",
                x: buttonRect.left,
                y: buttonRect.top,
                width: buttonRect.width,
                height: buttonRect.height
            });
        },

        _onExpandTransition = function() {
            if (_expandableContainer.offsetHeight === 0) {
                _headerContainer.classList.remove(EXPANDED_CONTAINER_CLASS_NAME);
                _expandableContainer.style.display = "none";
            }

            _expandableContainer.style.width = "auto";
            _expandableContainer.style.height = "auto";
        },
    
        /**
         Drag and Drop recipients
         Recipients can be reordered via drag and drop or moved between to, cc, and bcc lists
         Reordering is achieved by dragging a contact over another recipient (it will be moved before that recipient).
         Moving between lists is achieved by dropping a recipient to an empty part of recipient list (it will be moved to the end of that list).
         */
        _onRecipientListDragEnter = function(e) {
            e.preventDefault();
            e.stopPropagation();
            const draggedElement = document.querySelector("#" + _recipientDraggedId);
            this.insertBefore(draggedElement, this.lastElementChild);
        },
    
        _onRecipientListDragOver = function(e) {
            e.preventDefault();
            e.stopPropagation();
        },

        _setupDragAndDrop = function() {
            document.querySelectorAll("." + RECIPIENT_LIST_CLASS_NAME).forEach(ul => {
                ul.addEventListener("dragover", _onRecipientListDragOver);
                ul.addEventListener("dragenter", _onRecipientListDragEnter);
            });
        },
    
        _onRecipientDragStart = function(e) {
            _selectRecipientEl(this);
            e.dataTransfer.effectAllowed = 'move';
            _recipientDraggedId = this.id;
            // Required by drag and drop
            e.dataTransfer.setData("id", this.id);
        },
    
        _onRecipientDragEnter = function(e) {
            e.preventDefault();
            e.stopPropagation();
            const draggedElement = document.querySelector("#" + _recipientDraggedId);
            if (this === draggedElement) {
                return;
            }
            const recipientList = this.closest("." + RECIPIENT_LIST_CLASS_NAME);
            recipientList.insertBefore(draggedElement, this);
        },
    
        _onRecipientDragOver = function(e) {
            e.preventDefault();
            e.stopPropagation();
        },
        
        // MARK: - Public
        /* PUBLIC METHODS */

        insertSubject = function(text) {
            _subjectFieldInput.value = text;
            
            return true;
        },

        insertText = function(text) {
            var textDiv = document.createElement("div"),
                cursorPlaceholder = YRSEInputHandler.insertCursorPlaceholder();

            textDiv.innerHTML = text.replace(/(?:\r\n|\r|\n)/g, "<br>");

            if (cursorPlaceholder) {
                cursorPlaceholder.parentNode.replaceChild(textDiv, cursorPlaceholder);
            } else {
                textDiv.innerHTML += "<br><br>";
                _composeArea.appendChild(textDiv);
            }

            _linkify();
            
            return true;
        },

        insertImage = function(path, guid, inline, filename) {
            var img = new Image(),
                containerDiv = _createAttachmentContainer(path, guid, filename),
                overlayDiv = document.createElement("div"),
                cursorParentNode,
                cursorPlaceholder,
                lineBreak,
                label;

            overlayDiv.className = ATTACHMENT_OVERLAY_CLASS_NAME;
            img.src = path;
            containerDiv.appendChild(img);
            containerDiv.appendChild(overlayDiv);
            containerDiv.appendChild(_createMoreButton());

            if (inline) {
                img.style = "max-width:100%";
                containerDiv.contentEditable = false;
                // Insert image after cursor or before signature
                cursorPlaceholder = YRSEInputHandler.insertCursorPlaceholder();
                lineBreak = _createLineBreak();
                if (cursorPlaceholder) {
                    cursorParentNode = cursorPlaceholder.parentNode;
                    cursorParentNode.replaceChild(lineBreak, cursorPlaceholder);
                    cursorParentNode.insertBefore(containerDiv, lineBreak);
                } else if (_getSignatureContainer()) {
                    _composeArea.insertBefore(containerDiv, _signatureContainer);
                    _composeArea.insertBefore(lineBreak, _signatureContainer);
                } else {
                    _composeArea.appendChild(containerDiv);
                    _composeArea.appendChild(lineBreak);
                }
            } else {
                label = document.createElement("p");
                label.textContent = filename;
                containerDiv.appendChild(label);
                _attachmentArea.appendChild(containerDiv);
            }

            return path;
        },

        insertFile = function(path, guid, filename, className) {
            var containerDiv = _createAttachmentContainer(path, guid, filename),
                thumbDiv = document.createElement("div"),
                label;

            containerDiv.classList.add(FILE_ATTACHMENT_CLASS_NAME);
            thumbDiv.className = OUTLINE_ATTACHMENT_CLASS_NAME + " " + className;

            label = document.createElement("p");
            label.textContent = filename;

            containerDiv.appendChild(thumbDiv);
            containerDiv.appendChild(_createMoreButton());
            containerDiv.appendChild(label);
            _attachmentArea.appendChild(containerDiv);

            return guid;
        },

        insertSignature = function(signature) {
            if (signature && signature.trim().length === 0) {
                return;
            }

            _signatureContainer = document.createElement("div");
            _signatureContainer.id = SIGNATURE_ID;
            _signatureContainer.innerHTML = "<div><br></div><div>" + signature + "</div>";
            _composeArea.appendChild(_signatureContainer);
            
            return true;
        },

        insertContact = function(name, email) {
            if (!_focusedInput) {
                _postNativeMsg("error:no input focused");
                return;
            }

            var recipientHeader = _focusedInput.closest("." + RECIPIENT_HEADER_CLASS_NAME),
                recipientList = recipientHeader.querySelector("." + RECIPIENT_LIST_CLASS_NAME);
            recipientList.insertBefore(_createRecipientElement(name, email), recipientList.lastElementChild);
            _focusedInput.value = "";
            _postNativeMsg("contact:close");
            _postNativeMsg("event:log", {
                name: "recipient-added",
                fieldId: recipientHeader.id
            });
            _validateEmail();

            return name + ":" + email;
        },

        getMessageElements = function() {
            return {
                to: _getRecipients(TO_CONTAINER_ID),
                cc: _getRecipients(CC_CONTAINER_ID),
                bcc: _getRecipients(BCC_CONTAINER_ID),
                from: _getSenderEmail(),
                subject: _subjectFieldInput.value,
                body: _getTranslatedBody(),
                inlineGuids: _getGuids(_composeArea),
                outlineGuids: _getGuids(_attachmentArea)
            };
        },

        moveAttachmentInline = function(guid) {
            var attachmentDiv = _attachmentWithGuid(_attachmentArea, guid),
                path = attachmentDiv.getAttribute(DATA_PATH_ATTRIBUTE),
                filename = attachmentDiv.getAttribute(DATA_NAME_ATTRIBUTE);

            removeAttachment(guid);
            insertImage(path, guid, true, filename);
            
            return true;
        },

        moveAttachmentOutline = function(guid) {
            var attachmentDiv = _attachmentWithGuid(_composeArea, guid),
                path = attachmentDiv.getAttribute(DATA_PATH_ATTRIBUTE),
                filename = attachmentDiv.getAttribute(DATA_NAME_ATTRIBUTE);

            removeAttachment(guid);
            insertImage(path, guid, false, filename);
            
            return true;
        },

        removeAttachment = function(guid) {
            var attachmentDiv = document.querySelector("[" + DATA_GUID_ATTRIBUTE + "='" + guid + "']");

            if (attachmentDiv.closest("#" + COMPOSE_AREA_ID) && _isLineBreak(attachmentDiv.nextSibling)) {
                attachmentDiv.nextSibling.remove();
            }

            attachmentDiv.remove();
            
            return true;
        },

        insertSender = function(email) {
            var fromElement = _fromHeaderContainer.querySelector("." + FROM_ELEMENT_CLASS_NAME + " span"),
                originalSender = fromElement.textContent;

            fromElement.textContent = email;

            if (originalSender !== email) {
                _postNativeMsg("event:log", {
                    name: "sender-switched"
                });
            }
            
            return true;
        },

        focusRecipientField = function() {
            var input = document.querySelector("input[type=email]");

            if (input) {
                input.style.display = "";
                input.focus();
            }
            
            return true;
        };

    document.addEventListener("DOMContentLoaded", function(event) {
        _initElements();

        var recipientInputs = _headerContainer.querySelectorAll("." + RECIPIENT_HEADER_CLASS_NAME + " input");
        // Compose on mac does not register touch* events
        const touchEndEvent = YRSEConfig.isTouchInteraction ? "touchend" : "mouseup",
              touchStartEvent = YRSEConfig.isTouchInteraction ? "touchstart" : "mousedown";

        recipientInputs.forEach(function(el) {
            el.addEventListener("input", _onRecipientInput);
            el.addEventListener("keyup", _onRecipientKeyup);
            el.addEventListener("focus", _onRecipientFocus);
            el.addEventListener("blur", _onRecipientBlur);
            el.closest("." + RECIPIENT_LIST_CLASS_NAME).addEventListener(touchStartEvent, _onRecipientTouchStart);
        });

        _subjectFieldInput.addEventListener("focus", _onSubjectFocus);
        _subjectFieldInput.addEventListener("keyup", _onSubjectKeyup);
        _composeArea.addEventListener("input", _onMessageInput);
        _composeArea.addEventListener("keyup", _onMessageKeyup);
        _composeArea.addEventListener("focus", _onMessageFocus);
        _composeArea.addEventListener("paste", _onMessagePaste);
        _composeArea.addEventListener("blur", _onMessageBlur);

        _toHeaderContainer.querySelector("label").addEventListener("click", _onToLabelClick);
        _expandableContainer.addEventListener("transitionend", _onExpandTransition);
        document.addEventListener(touchEndEvent, _onDocumentTouchEnd);

        if (_fromHeaderContainer) {
            _fromHeaderContainer.querySelector("." + FROM_ELEMENT_CLASS_NAME).addEventListener("click", _onFromClick);
        }
        
        // Setup drag and drop
        _setupDragAndDrop();

        _postNativeMsg("dom:loaded");
    });
    
    // MARK: - Public methods
    // All methods below must return a value to avoid `Fatal error: Bug in WebKit: Received neither result or failure` when evaluated via `WKWebView.evaluateJavaScript()`
    return {
        insertSubject: insertSubject,
        insertText: insertText,
        insertImage: insertImage,
        insertFile: insertFile,
        insertSignature: insertSignature,
        insertContact: insertContact,
        getMessageElements: getMessageElements,
        moveAttachmentInline: moveAttachmentInline,
        moveAttachmentOutline: moveAttachmentOutline,
        removeAttachment: removeAttachment,
        insertSender: insertSender,
        focusRecipientField: focusRecipientField
    };
}());
