var YRSEInputHandler = (function() {
    var CURSOR_PADDING = 5,
        HEADER_ID = "yahoo__header",
        COMPOSE_AREA_ID = "yahoo__compose_area",
        CURSOR_PLACEHOLDER_CLASS_NAME = "cursor-placeholder",
        _headerDiv,

        /* PRIVATE */
        _getHeaderDiv = function() {
            if (_headerDiv) {
                return _headerDiv;
            }

            _headerDiv = document.getElementById(HEADER_ID);

            return _headerDiv;
        },

        _removeCursorPlaceholder = function() {
            document.querySelectorAll("." + CURSOR_PLACEHOLDER_CLASS_NAME).forEach(function(el) {
                el.remove();
            });
        },

        _getScrollHeights = function() {
            var cursorPlaceholder = insertCursorPlaceholder(),
                headerHeight = _getHeaderDiv().offsetHeight,
                scrollHeights = {};

            scrollHeights.top = headerHeight + cursorPlaceholder.offsetTop - CURSOR_PADDING;
            scrollHeights.bottom = scrollHeights.top + cursorPlaceholder.offsetHeight + CURSOR_PADDING * 2;
            _removeCursorPlaceholder();

            return scrollHeights;
        },

        _getPlaceholderSpan = function() {
            var span = document.createElement("span");
            span.className = CURSOR_PLACEHOLDER_CLASS_NAME;
            span.innerHTML = "&nbsp;";
            return span;
        },

        /* PUBLIC */
        getCursorScrollPositionIfNeeded = function() {
            if (!window.YRSETrueViewportHeight) {
                return null;
            }

            var scrollHeights = _getScrollHeights(),
                visibleTop = document.body.scrollTop,
                visibleBottom = visibleTop + window.YRSETrueViewportHeight;

            if (visibleBottom < scrollHeights.bottom) {
                return scrollHeights.bottom - window.YRSETrueViewportHeight;
            } else if (visibleTop > scrollHeights.top) {
                return scrollHeights.top;
            }

            return null;
        },

        insertCursorPlaceholder = function() {
            if (!document.activeElement || !document.activeElement.closest("#" + COMPOSE_AREA_ID)) {
                return null;
            }

            var range = window.getSelection().getRangeAt(0),
                cursorPlaceholder = _getPlaceholderSpan();

            range.deleteContents();
            range.insertNode(cursorPlaceholder);

            return cursorPlaceholder;
        };
    
    // MARK: - Public methods
    // All methods below must return a value to avoid `Fatal error: Bug in WebKit: Received neither result or failure` when evaluated via `WKWebView.evaluateJavaScript()`
    return {
        getCursorScrollPositionIfNeeded: getCursorScrollPositionIfNeeded,
        insertCursorPlaceholder: insertCursorPlaceholder
    };
}());
