var YRSELinkify = (function() {
    var URL_REGEX = /(http(s?):\/\/($|\S+)|www\.*.?\.($|\S+))/ig,

        /* PRIVATE METHODS */
        _createLinkNode = function(url) {
            var linkNode = document.createElement("a");
            linkNode.innerHTML = url;
            linkNode.href = url.toLowerCase().indexOf("http") === 0 ? url : "http://" + url;
            linkNode.rel = "nofollow";
            linkNode.target = "_blank";
            return linkNode;
        },

        /* PUBLIC METHODS */
        linkify = function(element) {
            var childNodes = element.childNodes,
                linkified = false,
                lastIndex = 0,
                originalContent,
                rightContent,
                leftContent,
                nodeName,
                textNode,
                result;

            childNodes.forEach(function(el) {
                // Process text nodes
                if (el.nodeType === 3) {
                    originalContent = el.textContent;
                    while ((result = URL_REGEX.exec(originalContent)) !== null) {
                        leftContent = originalContent.substring(lastIndex, URL_REGEX.lastIndex - result[0].length);
                        if (leftContent) {
                            textNode = document.createTextNode(leftContent);
                            element.insertBefore(textNode, el);
                        }
                        element.insertBefore(_createLinkNode(result[0]), el);
                        lastIndex = URL_REGEX.lastIndex;
                        rightContent = originalContent.substring(URL_REGEX.lastIndex);
                        linkified = true;
                    }

                    if (rightContent) {
                        textNode = document.createTextNode(rightContent);
                        element.insertBefore(textNode, el);
                    }

                    if (linkified) {
                        el.remove();
                    }
                }
                // Descend divs and spans
                else {
                    nodeName = el.nodeName.toLowerCase();
                    if (nodeName === "span" || nodeName === "div") {
                        linkify(el);
                    }
                }
            });
            
            return true;
        };
    
    // MARK: - Public methods
    // All methods below must return a value to avoid `Fatal error: Bug in WebKit: Received neither result or failure` when evaluated via `WKWebView.evaluateJavaScript()`
    return {
        linkify: linkify
    };
}());
