/**
 * YMUWebViewAttachmentHandler provides JavaScript methods and static variables for handling attachments in web views.
 */
var webViewAttachmentHandler = (function () {

    /** Saved cursor selection range (start to end). */
    var cursorSpanID = 'yahoo-rte-cursor-span',
    selectionRange,
    attachmentMenuHorizontalOffset = 32,
    _isExtensionActive = false,
    attachments_area_id = 'yahoo__attachments',

    // MARK: Private methods
    /**
     * Returns true if node is <BR> element.
     */
    isBR = function(node) {
        return node && node.tagName === "BR";
    },
    /**
     * Returns compose area of message body.
     */
    getComposeArea = function() {
        return document.querySelector("[contenteditable='true']");
    },
    /**
     * Adds a dummy invisible cursorSpan at the current Selection
     * Also saves the current cursor selection in a static variable for later use at insertNodeAtCursor.
     */
    addCursorSpan = function () {
        //current selection may be null
        var selection,cursorSpan, composeArea;

        selection = window.getSelection();
        cursorSpan = document.getElementById(cursorSpanID);

        if (cursorSpan && !selection.rangeCount) {
            // we want to preserve the previously inserted cursor position
            return false;
        }

        composeArea = getComposeArea();
        composeArea.focus();

        if (selection.getRangeAt) {
            if (cursorSpan) {
                cursorSpan.remove();
            }

            cursorSpan = document.createElement('span');
            cursorSpan.id = cursorSpanID;

            // insert Cursor span and remove cursor span from current selection
            try {
                selectionRange = selection.getRangeAt(0);
            } catch (err) {
                return null;
            }

            selectionRange.insertNode(cursorSpan);
            selectionRange.setStartAfter(cursorSpan);
        }
        
        return true;
    },
    
    removeCursorSpan = function () {
        cursorSpan = document.getElementById(cursorSpanID);
        if (cursorSpan) {
            cursorSpan.remove();
        }
        
        return true;
    },
    /**
     * Sets whether app extension is currently active
     *
     * @param {Bool} active - Whether extension is active
     */
    setExtensionActive = function(active) {
        _isExtensionActive = active;
        
        return true;
    },
    restoreFocusAtLastKnownPosition = function(alignToTop) {
        var sel, range, cusorSpan, composeArea;

        composeArea = getComposeArea();
        composeArea.focus();
        cursorSpan = document.getElementById(cursorSpanID);

        sel = window.getSelection();
        range = document.createRange();
        range.selectNodeContents(cursorSpan);
        sel.removeAllRanges();
        sel.addRange(range);
        cursorSpan.scrollIntoView(alignToTop);
        
        return true;
    },
    getOutlineFileIdentifiers = function() {
        var i, list, files, div, guid, downloadSrc, fileName, fileSize;
        files = document.querySelectorAll("div[class^=attachment][data-guid]");
        list = [];
        for (i = 0; i < files.length; i++) {
            div = files[i];
            guid = div.getAttribute("data-guid");
            downloadSrc = div.getAttribute("download-src");
            fileName = div.getAttribute("filename");
            fileSize = div.getAttribute("filesize");
            var startsWithRegex = /^webkit-fake-url/;
            if (downloadSrc && fileName && fileSize && downloadSrc != guid && !startsWithRegex.test(downloadSrc)) continue;
            list.push(guid);
        }
        return list.join();
    },
    getOutlineRemoteAttachments = function() {
        var i, list, files, div, guid, downloadSrc, fileName, fileSize, partID, refID, contentType;
        files = document.querySelectorAll("div[class^=attachment][data-guid]");
        list = [];
        for (i = 0; i < files.length; i++) {
            div = files[i];
            guid = div.getAttribute("data-guid");
            downloadSrc = div.getAttribute("download-src");
            fileName = div.getAttribute("filename");
            fileSize = div.getAttribute("filesize");
            partID = div.getAttribute("data-partid") || "";
            refID = div.getAttribute("data-refid") || "" ;
            contentType = div.getAttribute("data-contenttype") || "";

            if (downloadSrc && fileName && fileSize) {
                list.push("ios://attachment?guid=" + encodeURIComponent(guid) + "&url=" + encodeURIComponent(downloadSrc) + "&filename=" + encodeURIComponent(fileName) + "&filesize=" + encodeURIComponent(fileSize) + "&partID=" +  encodeURIComponent(partID) + "&refID=" + encodeURIComponent(refID) + "&contentType=" + encodeURIComponent(contentType));
            }
        }
        return list.join("\u2028");
    },
    getImagesCountWithWebkitFakeURL = function() {
        var webKitImages = document.querySelectorAll('img[src^=webkit-fake-url]:not([data-guid])');
        return webKitImages.length;
    },
    updateImagesWithWebkitFakeURL = function(fileNamesString, fileSizesString) {
        var i, webKitImages, fileNames, webKitImage, fileName, fileSize, parent, container, fileSizes;
        webKitImages = document.querySelectorAll("img[src^=webkit-fake-url]:not([data-guid])");
        fileNames = fileNamesString.split(",");
        fileSizes = fileSizesString.split(",");
        if (fileNames.length === webKitImages.length) {
            for (i = 0; i < webKitImages.length; i++) {
                webKitImage = webKitImages[i];
                fileName = fileNames[i];
                fileSize = fileSizes[i];
                parent = webKitImage.parentNode;
                container = getContainerForInlineAttachment(webKitImage.src, fileName, fileName.split("_")[1], fileSize, "image");
                parent.replaceChild(container, webKitImage);
            }
        }
        return webKitImages.length;
    },
    getContainerForInlineAttachment = function(imageSource, dataId, fileName, fileSize, attachmentType) {
        return getContainerForInlineAttachmentWithReference(imageSource, dataId, fileName, fileSize, "", "", attachmentType);
    },
    getContainerForInlineAttachmentWithReference = function(imageSource, dataId, fileName, fileSize, partID, refID, attachmentType, maxWidth) {
        var container = Template.loadTemplate("attachment_inline", {
            imageSource: imageSource,
            dataId: dataId,
            fileName: fileName,
            fileSize: fileSize,
            partID: partID,
            refID: refID,
            attachmentType: attachmentType,
            maxWidth: maxWidth,
            attachmentMoreMenu: LocalizedString.attachment_more_menu
        });
                                
        if (!_isExtensionActive) {
            container.querySelector('img').addEventListener("load",  function(event) {
                InputHandler.scrollIfNeeded();
            });
        }
                                
        return container;
    },
    getContainerForLinkAttachment = function(url, dataId, title, subtitle, thumbnailURL) {
        var container = Template.loadTemplate("attachment_link", {
            url: url,
            dataId: dataId,
            title: title,
            subtitle: subtitle,
            thumbnailURL: thumbnailURL
        });
        return container;
    },
    getContainerForLinkWithUrl = function(url, dataId, attachmentType, maxWidth) {
        var container = Template.loadTemplate("attachment_link_url", {
            url: url,
            dataId: dataId,
            attachmentType: attachmentType,
            maxWidth: maxWidth,
            attachmentMoreMenu: LocalizedString.attachment_more_menu
        });
                                
        if (!_isExtensionActive) {
            container.querySelector('img').addEventListener("load",  function(event) {
                InputHandler.scrollIfNeeded();
            });
        }
        
        return container;
    },

    getContainerForTumblrGifAttribution = function(pageName, pageUrl, dataId) {
        var container = Template.loadTemplate("attachment_tumblr_attribution", {
            pageName: pageName,
            pageUrl: pageUrl,
            dataId: dataId
        });
                                
        if (!_isExtensionActive) {
            container.querySelector('img').addEventListener("load",  function(event) {
                InputHandler.scrollIfNeeded();
            });
        }
        
        return container;
    },
                                
    getContainerForTenorGifAttribution = function(pageUrl, dataId) {
        var container = Template.loadTemplate("attachment_tenor_attribution", {
                                          pageUrl: pageUrl,
                                          dataId: dataId
                                          });
    
        if (!_isExtensionActive) {
        container.querySelector('img').addEventListener("load",  function(event) {
                                                    InputHandler.scrollIfNeeded();
                                                    });
        }
    
        return container;
    },

    getContainerForGifWithUrl = function(url, dataId, pageName, pageUrl, source, attachmentType, maxWidth) {
        var gifContainer, wrapperContainer, attribution;
        var isTumblr = (pageName && (source.localeCompare("www@tumblr") === 0));
        var isTenor = (source.localeCompare("www@tenor") === 0);
                                
        gifContainer = getContainerForLinkWithUrl(url, dataId, attachmentType, maxWidth);

        if (isTumblr || isTenor) {
            attribution = isTenor ? getContainerForTenorGifAttribution(pageUrl, dataId) : getContainerForTumblrGifAttribution(pageName, pageUrl, dataId);
            gifContainer.className = "gif-wrapper";
            wrapperContainer = document.createElement("div");
            wrapperContainer.className = "card-wrapper-link";
            wrapperContainer.contentEditable = false;
            wrapperContainer.setAttribute("data-guid", dataId);
            wrapperContainer.appendChild(gifContainer);
            wrapperContainer.appendChild(attribution);
            return wrapperContainer;
        } else {
            return gifContainer;
        }
    },

    insertInlineAttachment = function(imageSource, dataId, fileName, fileSize, attachmentType) {
        return insertInlineAttachmentWithReference(imageSource, dataId, fileName, fileSize, "", "", attachmentType);
    },
    /**
     * Inserts inline attachment into document.
     * @param {string} imageSource
     * @param {string} dataId
     * @param {string} fileName
     * @param {Number} fileSize
     * @param {string} attachmentType
     * @param {string} maxWidth
     * @return {Number} container thumbnail
     */
    insertInlineAttachmentWithReference = function(imageSource, dataId, fileName, fileSize, partID, refID, attachmentType, maxWidth) {
        var container;
        addCursorSpan();
        if (imageSource) {
            container = getContainerForInlineAttachmentWithReference(imageSource, dataId, fileName, fileSize, partID, refID, attachmentType, maxWidth);
            container.addEventListener("touchend",  function(event) {
                event.preventDefault();
            });
            insertNodeAtCursor(container);
            return container;
        }
    },
    updateInlineAttachmentWithReference = function(guid, imageSource) {
        var parent, node;
        parent = document.querySelector('div[data-guid="'+guid+'"]');
        parent.setAttribute('data-image-source', imageSource);
        node = parent.getElementsByTagName('img');
        node[0].src = imageSource;
        
        return true;
    },
    insertText = function(text, enhanceLinks, focus) {
        var composeArea = getComposeArea(),
            textElement = document.createElement("div"),
            range = document.createRange(),
            selection = window.getSelection();
        textElement.innerHTML = text.replace(/(?:\r\n|\r|\n)/g, '<br>');
        composeArea.insertBefore(textElement, composeArea.firstChild);
        range.setStart(textElement.nextSibling, 0);
        range.setEnd(textElement.nextSibling, 0);
        selection.removeAllRanges();
        selection.addRange(range);
        if (enhanceLinks) {
            LinkEnhancr.enhanceLinks(composeArea);
        }
        if (focus) {
            composeArea.focus();
        }
        
        return true;
    },
    insertTextAtCursor = function(text, enhanceLinks, focus, isURL) {
        var composeArea = getComposeArea(), sel, range, textElement;
        if(!isURL) {
            sel = window.getSelection();
            if (sel.getRangeAt && sel.rangeCount) {
                range = sel.getRangeAt(0);
                range.deleteContents();
                range.insertNode( document.createTextNode(text) );
            }
        } else {
            textElement = document.createElement("div");
            textElement.innerHTML = text.replace(/(?:\r\n|\r|\n)/g, '<br>');
            insertNodeAtCursor(textElement);
        }
                                
        if (enhanceLinks) {
            LinkEnhancr.enhanceLinks(composeArea);
        }
        if (focus) {
            composeArea.focus();
        }
        
        return true;
    },
    insertImage = function(imageSource, dataId, fileName, fileSize, partID, refID, inline, maxWidth) {
        if (inline) {
            insertInlineAttachmentWithReference(imageSource, dataId, fileName, fileSize, partID, refID, "image", maxWidth);
        }
        else {
            getComposeArea().focus();
            insertOutlineAttachmentWithReference(imageSource, dataId, fileName, fileSize, partID, refID, "image", maxWidth);
        }
        
        return true;
    },
    insertVideo = function(imageSource, dataId, fileName, fileSize, inline) {
        insertVideoWithReference(imageSource, dataId, fileName, fileSize, "", "", inline);
        
        return true;
    },
    insertVideoWithReference = function(imageSource, dataId, fileName, fileSize, partID, refID, inline) {
        var playButton, button, container;
        playButton = function(className) {
            button = document.createElement("button");
            button.className = className;
            button.contentEditable = false;
            button.setAttribute("data-guid", dataId);
            return button;
        };

        if (inline) {
            container = insertInlineAttachmentWithReference(imageSource, dataId, fileName, fileSize, partID, refID, "video");
            container.appendChild(playButton("play"));
        }
        else {
            getComposeArea().focus();
            insertOutlineAttachmentWithReference(imageSource, dataId, fileName, fileSize, partID, refID, "video");
        }
        
        return true;
    },
    insertAudioWithReference = function(imageSource, dataId, fileName, fileSize, partID, refID, inline) {
        var playButton, button, container;
        playButton = function(className) {
            button = document.createElement("button");
            button.className = className;
            button.contentEditable = false;
            button.setAttribute("data-guid", dataId);
            return button;
        };

        if (inline) {
            container = insertInlineAttachmentWithReference(imageSource, dataId, fileName, fileSize, partID, refID, "audio");
            container.appendChild(playButton("play"));
        }
        else {
            getComposeArea().focus();
            insertOutlineAttachmentWithReference(imageSource, dataId, fileName, fileSize, partID, refID, "audio");
        }
        
        return true;
    },
    insertDocument = function(imageSource, dataId, fileName, fileSize, partID, refID, inline, iconType, contentType) {
        if (inline) {
            insertInlineAttachmentWithReference(imageSource, dataId, fileName, fileSize, partID, refID, "document");
        }
        else {
            getComposeArea().focus();
            insertOutlineAttachmentWithReference(imageSource, dataId, fileName, fileSize, partID, refID, "document", "", iconType, contentType);
        }
        
        return true;
    },
    insertLink = function(url, dataId, title, subtitle, thumbnailURL) {
        var container;

        if (thumbnailURL && getComposeArea()) {
            container = getContainerForLinkAttachment(url, dataId, title, subtitle, thumbnailURL);
            insertNodeAtCursor(container);
            return container;
        }
        
        return false;
    },
    insertGif = function(url, dataId, pageName, pageUrl, source, maxWidth) {
        var container;
        if (url) {
            container = getContainerForGifWithUrl(url, dataId, pageName, pageUrl, source, "link", maxWidth);
            insertNodeAtCursor(container);
            return container;
        }
        
        return false;
    },
    /**
     * Shows outline attachment drawer when outline attachment thumbnail pressed (even around ...). Calls
     *   "ios://show-menu/<data-guid>/<CGPoint>?url=<download-src>&filename=<filename>&filesize=<filesize>&type=<attachment-type>"
     */
    attachmentTapped = function(div, event, attachmentType) {
        var target, bodyRect, elemRect, x, y;
        event.preventDefault();
        event.stopPropagation();
        target = event.currentTarget;
        bodyRect = document.body.getBoundingClientRect();
        elemRect = target.getBoundingClientRect();
        x = elemRect.left + elemRect.width - attachmentMenuHorizontalOffset;
        y = elemRect.top - bodyRect.top;
        window.location = "ios://show-menu/" + encodeURIComponent(target.getAttribute("data-guid")) + "/" + x + ":" + y + "?url=" + encodeURIComponent(target.getAttribute("download-src")) + "&filename=" + encodeURIComponent(target.getAttribute("filename")) + "&filesize=" + encodeURIComponent(target.getAttribute("filesize")) + "&type=" + attachmentType;
        // MI-16814: iOS 11 fix to prevent bringing up keyboard which overlaps menu
        setTimeout(function() {
            document.activeElement && document.activeElement.blur();
        }, 100);
        
        return true;
    },
    /**
     * Shows horizontal inline attachment menu when image gets pressed. Calls
     *   "ios://show-menu/<data-guid>/<CGPoint>?type=<attachment-type>"
     */
    moreAttachmentButtonPressed = function(event) {
        var target, bodyRect, elemRect, x, y, attachmentType, borderOffset, changedTouch;

        if (event instanceof Event) {
            event.preventDefault();
            event.stopPropagation();
        }
        if (event.target) {
            target = event.target;
        } else {
            return false;
        }
        
        if (event.changedTouches) {
            changedTouch = event.changedTouches[0];
            if (changedTouch && changedTouch.target.tagName != "BUTTON") {
                return false;
            }
        }
        
        attachmentType = target.getAttribute('data-type') || target.parentNode.getAttribute('data-type');
        borderOffset = attachmentType ? 1 : 0;
        bodyRect = document.body.getBoundingClientRect();
        elemRect = target.getBoundingClientRect();
        x = elemRect.right - attachmentMenuHorizontalOffset + borderOffset;
        y = elemRect.top - bodyRect.top - borderOffset;

        window.location = "ios://show-menu/" + encodeURIComponent(target.getAttribute("data-guid")) + "/" + x + ":" + y + "?type=" + attachmentType;
        
        return true;
    },
    /**
     * Converts inline attachment to outline attachment.
     * @param {string} dataId
     */
    convertInlineAttachmentToOutline = function(dataId) {
        var i, attachments, attachment, imageSource, fileName, type, fileSize, partID, refID;
        attachments = document.querySelectorAll("div[class^=inline-attachment-container][data-guid]");
        for (i = 0; i < attachments.length; i++) {
            attachment = attachments[i];
            if (attachment.getAttribute("data-guid") === dataId) {
                imageSource = attachment.getAttribute("data-image-source");
                fileName = attachment.getAttribute("data-filename");
                type = attachment.getAttribute("data-type");
                fileSize = attachment.getAttribute("data-filesize");
                partID = attachment.getAttribute("data-partid") || "";
                refID = attachment.getAttribute("data-refid") || "";
                maxWidth = attachment.getAttribute("max-width") || "";
                attachment.parentElement.removeChild(attachment);
                insertOutlineAttachmentWithReference(imageSource, dataId, fileName, fileSize, partID, refID, type, maxWidth);
            }
        }
        
        return true;
    },
    /**
     * Converts outline attachment to inline attachment. "attachment " + type
     * @param {string} dataId
     */
    convertOutlineAttachmentToInline = function(dataId) {
        var i, attachments, downloadSource, fileName, type, fileSize, partID, refID;
        attachments = document.querySelectorAll("div[class^=attachment ][data-guid]");
        for (i = 0; i < attachments.length; i++) {
            attachment = attachments[i];
            if (attachment.getAttribute("data-guid") === dataId) {
                downloadSource = attachment.getAttribute("download-src");
                fileName = attachment.getAttribute("filename");
                type = attachment.className.split(" ")[1];
                fileSize = attachment.getAttribute("filesize");
                // partID / refID could be null
                partID = attachment.getAttribute("data-partid") || "";
                refID = attachment.getAttribute("data-refid")|| "";
                maxWidth = attachment.getAttribute("max-width")|| "";
                attachment.parentElement.removeChild(attachment);
                insertInlineAttachmentWithReference(downloadSource, dataId, fileName, fileSize, partID, refID, type, maxWidth);
            }
        }
        
        return true;
    },
    removeInlineAttachment = function(dataId) {
        var i, attachments, attachment;
        attachments = document.querySelectorAll("div[class^=inline-attachment-container][data-guid]");
        for (i = 0; i < attachments.length; i++) {
            attachment = attachments[i];
            if (attachment.getAttribute("data-guid").indexOf(dataId) >= 0) {
                attachment.remove();
            }
        }
        
        return true;
    },
    updateLinkAttachment = function(guid, url) {
        var attachment = document.querySelector('a[data-guid*="'+guid+'"]');
        if (attachment) attachment.setAttribute("href", url);
        
        return true;
    },
    removeLinkAttachment = function(dataId) {
        var i, attachments, attachment;
        attachments = document.querySelectorAll("div[class^=card-wrapper-link][data-guid], a[class^=card-wrapper-link][data-guid], div[class^=gif-wrapper][data-guid]");
        for (i = 0; i < attachments.length; i++) {
            attachment = attachments[i];
            if (attachment.getAttribute("data-guid").indexOf(dataId) >= 0) {
                attachment.parentElement.removeChild(attachment);
            }
        }
        
        return true;
    },
    removeOutlineAttachment = function(dataId) {
        var i, attachments, attachment;
        bodyUpdated();
        attachments = document.querySelectorAll("div[class^=attachment][data-guid]");
        for (i = 0; i < attachments.length; i++) {
            attachment = attachments[i];
            if (attachment.getAttribute("data-guid").indexOf(dataId) >= 0) {
                attachment.parentElement.removeChild(attachment);
            }
        }
        window.location = "mailsdk://removeAttachment";
        
        return true;
    },
    /**
     * Inserts outline attachment.
     * @param {string} downloadSrc
     * @param {string} dataId
     * @param {string} fileName
     * @param {Number} fileSize
     * @param {string} attachmentType
     * @param {string} maxWidth
     * @param {string} iconType
     * @param {string} contentType
     */
    insertOutlineAttachmentWithReference = function(downloadSrc, dataId, fileName, fileSize, partID, refID, attachmentType, maxWidth, iconType, contentType) {
        var container = Template.loadTemplate("attachment_outline", {
            downloadSrc: downloadSrc,
            dataId: dataId,
            fileName: fileName,
            fileSize: fileSize,
            partID: partID,
            refID: refID,
            attachmentType: attachmentType,
            maxWidth: maxWidth,
            iconType: iconType,
            contentType: contentType,
            attachmentTypeIsImage: attachmentType === "image",
            thumbnailPlaceholderType: function() {
                if (attachmentType === "video" || attachmentType === "audio") {
                    return attachmentType;
                } else if (attachmentType === "document" && iconType) {
                    return iconType;
                }
                return "file";
            }(),
            attachmentMoreMenu: LocalizedString.attachment_more_menu
        });
        container.addEventListener("click",  function(event) {
            webViewAttachmentHandler.attachmentTapped(this, event, attachmentType);
        });
        document.getElementById("yahoo__attachments").appendChild(container);
    },
    removeAttachment = function(button, event, dataId) {
        var i, attachments, attachment;
        if (event) {
            event.preventDefault();
            event.stopPropagation();
            bodyUpdated();
        }
        attachments = document.querySelectorAll("div[class^=attachment][data-guid]");
        for (i = 0; i < attachments.length; i++) {
            attachment = attachments[i];
            if (attachment.getAttribute("data-guid").indexOf(dataId) >= 0) {
                attachment.parentElement.removeChild(attachment);
            }
        }
        if (event) {
            window.location = "mailsdk://removeAttachment";
        }
        
        return true;
    },
    removeAttachments = function(GUIDs) {
        var i, guid;
        for (i = 0; i < GUIDs.length; i++) {
            guid = GUIDs[i];
            removeInlineAttachment(guid);
            removeAttachment(null, null, guid);
        }
        
        return true;
    },
    insertAfter = function(newNode, referenceNode) {
        referenceNode.parentNode.insertBefore(newNode, referenceNode.nextSibling);
    },
    getLineBreakNode = function() {
        var containerNode = document.createElement("div");
        containerNode.appendChild(document.createElement("br"));
        return containerNode;
    },
    isNodeContentEditable = function(node) {
        return node.getAttribute("contenteditable") !== "false";
    },
    isNodeEmptyText = function(node) {
        if (node && node.nodeType === 3 && node.length === 0) {
            return true;
        }
        return false;
    },
    insertNodeAtCursor = function(newNode) {
        var range, composeArea, previousElement, nextNode, parentNode;
        cursorSpan = document.getElementById(cursorSpanID);

        if (!cursorSpan) {
            addCursorSpan();
            cursorSpan = document.getElementById(cursorSpanID);
        }

        if (selectionRange) {
            selectionRange.deleteContents();
            selectionRange = null;
        }

        if (cursorSpan) {
            parentNode = cursorSpan.parentNode;
            parentNode.insertBefore(newNode, cursorSpan);
            parentNode.insertBefore(getLineBreakNode(), cursorSpan);
        } else {
            // No cursor span found, so place at end of message
            parentNode = getComposeArea();
            parentNode.appendChild(newNode);
            parentNode.appendChild(getLineBreakNode());
        }

        // if the new node isn't content editable, it must have a previous non-empty node to be deleteable
        if (!isNodeContentEditable(newNode) && (!newNode.previousSibling || isNodeEmptyText(newNode.previousSibling))) {
            parentNode.insertBefore(getLineBreakNode(), newNode);
        }
    },

    moveCaretToPoint = function(point) {
        var range, selection;
        if (document.caretRangeFromPoint) {
            range = document.caretRangeFromPoint(point.x, point.y);
            if (range) {
                selection = window.getSelection();  //get the selection object (allows you to change selection)
                selection.removeAllRanges();        //remove any selections already made
                selection.addRange(range);
            }
        }
        
        return true;
    },
                                
    isComposeAreaAtPoint = function(point) {
        var el, composeArea;
        if (document.elementFromPoint) {
            el = document.elementFromPoint(point.x, point.y);
            composeArea = getComposeArea();
            return composeArea.contains && composeArea.contains(el);
        }
        return false;
    },

    _pastePasteboard = function() {
        window.location = "ios://paste-pasteboard/";
    },
    
    _pasteMacPasteboard = function() {
        window.location = "ios://paste-mac-pasteboard";
    }
    
    _pasteImageOnLoad = function(type, dataURI) {
        window.location = "ios://paste-image/?type=" + encodeURIComponent(type) + "&dataURI=" + encodeURIComponent(dataURI);
    },

    _pasteImage = function(item) {
        var blob = item.getAsFile(),
            reader = new FileReader();

        reader.onload = function(event) {
            setTimeout(_pasteImageOnLoad, 0, item.type, event.target.result);
        };

        reader.readAsDataURL(blob);
    },

    handlePasteEvent = function(event) {
        // MI-23685: Special-handling to capture case when pasted image results in pasted filepath on M1 Mac
        if (!isTouchInteraction && event.clipboardData.types.length === 0) {
            _pasteMacPasteboard();
            event.preventDefault();
            return false;
        }
        
        var items = event.clipboardData.items,
            types = event.clipboardData.types,
            length,
            hasImgData,
            index,
            item;

        if (items) {
            length = items.length;
        } else if (types) {
            _pastePasteboard();
            event.preventDefault();
            return false;
        }

        for (index = 0; index < length; index++) {
            item = items[index];
            if (item.kind === 'file' && item.type.match(/^image\//)) {
                hasImgData = true;
                _pasteImage(item);
            }
        }

        if (hasImgData) {
            event.preventDefault();
        }
        
        return true;
    },
    
    focusAndAddCursorSpan = function() {
        document.getElementById('yahoo__compose_area').focus();
        addCursorSpan();
        
        return true;
    };
    
    // MARK: - Public methods
    // All methods below must return a value to avoid `Fatal error: Bug in WebKit: Received neither result or failure` when evaluated via `WKWebView.evaluateJavaScript()`
    return {

    getOutlineFileIdentifiers: getOutlineFileIdentifiers,

    getOutlineRemoteAttachments: getOutlineRemoteAttachments,

    getImagesCountWithWebkitFakeURL: getImagesCountWithWebkitFakeURL,

    updateImagesWithWebkitFakeURL: updateImagesWithWebkitFakeURL,

    updateInlineAttachmentWithReference: updateInlineAttachmentWithReference,

    insertText: insertText,

    insertTextAtCursor: insertTextAtCursor,
    
    insertImage: insertImage,

    insertVideo: insertVideo,

    insertVideoWithReference: insertVideoWithReference,
        
    insertAudioWithReference: insertAudioWithReference,

    insertDocument: insertDocument,

    insertLink: insertLink,

    insertGif: insertGif,

    attachmentTapped: attachmentTapped,
                                
    setExtensionActive: setExtensionActive,
    
    restoreFocusAtLastKnownPosition: restoreFocusAtLastKnownPosition,

    moreAttachmentButtonPressed: moreAttachmentButtonPressed,

    convertInlineAttachmentToOutline: convertInlineAttachmentToOutline,

    convertOutlineAttachmentToInline: convertOutlineAttachmentToInline,

    removeInlineAttachment: removeInlineAttachment,

    removeOutlineAttachment: removeOutlineAttachment,

    removeLinkAttachment: removeLinkAttachment,

    updateLinkAttachment: updateLinkAttachment,

    removeAttachment: removeAttachment,

    addCursorSpan: addCursorSpan,

    removeAttachments: removeAttachments,
    
    moveCaretToPoint: moveCaretToPoint,
                                
    removeCursorSpan: removeCursorSpan,
    
    isComposeAreaAtPoint: isComposeAreaAtPoint,

    handlePasteEvent: handlePasteEvent,
        
    focusAndAddCursorSpan: focusAndAddCursorSpan
    };

}());
