var MessageComposeController = (function() {

var editingDidBegin = false;

var editableElement = function() {
    var bodyDiv = document.getElementById('yahoo__compose_area');
    return bodyDiv;
}

var bodyFocused = function (event) {
    if (event) {
        var elementClassList = event.target.classList;
        var classesToSkip = ['remove', 'more', 'play', 'play-small'];

        if (elementClassList && classesToSkip.some(className => elementClassList.contains(className))) {
            return;
        }
    }
    beginEditing();
}

var bodyUpdated = function (event) {
    InputHandler.scrollEnabled = true;
    if (editingDidBegin) {
        window.location = 'mailsdk:editingDidBegin';
        editingDidBegin = false;
    } else {
        window.location = 'mailsdk:contentDidChange';
    }
}

var focusInComposeArea = function() {
    document.querySelector("[contenteditable='true']").focus();
    
    return true;
}

var beginEditing = function(triggersContentDidTap) {
    InputHandler.scrollEnabled = true;
    const composeArea = editableElement();
    composeArea.focus();

    // MI-22478: On mac, initial focus in compose area doesn't trigger the formatting toolbar.  Blur and re-focus so that it does.  Use setTimeout with minimal delay to not block the main thread and allow blur and focus to register.
    if (!isTouchInteraction && !editingDidBegin) {
        setTimeout(() => {
            composeArea.blur();
            setTimeout(() => composeArea.focus())
        });
    }
    
    editingDidBegin = true;
    var placeholder = document.getElementById('placeholder');
    if (placeholder) {
        placeholder.remove();
    }
    if (triggersContentDidTap) {
        window.location = 'mailsdk:contentDidTap';
    }
    
    return true;
}

var updateFontSize = function(fontSize) {
    document.body.style.fontSize = fontSize;
    
    return true;
}

var extractContent = function() {
    return editableElement().innerHTML;
}

var interceptFocus = function(intercept) {
    document.body.removeEventListener('focus', bodyFocused, true);
    if (intercept) {
        document.body.addEventListener('focus', bodyFocused, true);
    }

    return true;
}

var pasteEventListener = function(event) {
    LinkEnhancr.handlePasteEvent(event);
    webViewAttachmentHandler.handlePasteEvent(event);
    if (!event.defaultPrevented) { // paste not handled by LinkEnhancr or webViewAttachmentHandler
        YMColorInverter.messageComposeLightModeHandlePaste();
    }
    setTimeout(function() {
        bodyUpdated();
    }, 0);
};

var interceptPaste = function(intercept) {
    document.body.removeEventListener('paste', pasteEventListener);
    if (intercept) {
        document.body.addEventListener('paste', pasteEventListener);
    }
}

document.addEventListener("DOMContentLoaded", function(event) {
    interceptFocus(true);
                          
    editableElement().addEventListener('input',  function(e) {
        // MI-16714: prevent clearing of text when using ios toolbar format options
        if (e.inputType.startsWith("format")) {
            return;
        }
                                       
        InputHandler.scrollIfNeeded();
        bodyUpdated();
    });
    
    // Compose on mac does not register touch* events
    const menuTriggerEvent = isTouchInteraction ? "touchend" : "mouseup";
    
    editableElement().addEventListener(menuTriggerEvent, function(touchEvent) {
        var targetNode = touchEvent.target,
            dataAction = targetNode.getAttribute('data-action');

        if (dataAction) {
            switch(dataAction) {
            case 'moreMenuInlineAttachment':
                webViewAttachmentHandler.moreAttachmentButtonPressed(touchEvent);
                break;
            }
        }
            
        if (touchEvent.target.className.match(/(.*)close-button(?!\S)/)) {
            var target = getAncestor(touchEvent.target, "card-wrapper-link");
            if (target) {
                touchEvent.preventDefault();
                target.parentElement.removeChild(target);
                bodyUpdated(event);
            }
        }
    });
    document.body.addEventListener('keypress', function (e) {
        if (e.keyCode === 13 || e.keyCode === 32) {
            LinkEnhancr.needsReturn = e.keyCode === 13;
            setTimeout(function() {
                LinkEnhancr.enhanceLinks(editableElement());
            }, 100);
        }
    });
    interceptPaste(true);
    LinkEnhancr.prepareLinkEnhancrCards();
    links = document.querySelectorAll('[class*="card-wrapper-link"]');
    for (i in links) {
        links[i].contentEditable = false;
    }
});

function getAncestor(node, className) {
    while (node.id !== document) {
        if (node.className.match(new RegExp('(.*)' + className))) {
            return node;
        }
        node = node.parentNode;
    }

    return null;
};
    
    // MARK: - Public methods
    // Expose to window as it's used externally by stationery.js
    window.bodyUpdated = bodyUpdated;
    
    // All methods below must return a value to avoid `Fatal error: Bug in WebKit: Received neither result or failure` when evaluated via `WKWebView.evaluateJavaScript()`
    return {
        editableElement: editableElement,
        focusInComposeArea: focusInComposeArea,
        beginEditing: beginEditing,
        updateFontSize: updateFontSize,
        extractContent: extractContent,
        interceptFocus: interceptFocus
    };
    
}());
