var webViewDragAndDropHandler = (function() {
    var ATTACHMENTS_AREA_ID = "yahoo__attachments",
        INLINE_ATTACHMENT_CONTAINER_ID = "inline-attachment-container",
        CUSTOM_INLINE_IMAGE = "custom-inline-image",
        CUSTOM_INLINE_VIDEO = "custom-inline-video",
        CUSTOM_INLINE_DOCUMENT = "custom-inline-document",
        CUSTOM_INLINE_AUDIO = "custom-inline-audio",
        _attachmentArea;

    dragItemAtPoint = function(point) {
            var el, node, thumbnailData, json, imageNodes,
                downloadSrcNode, filenameNode, filesizeNode, altNode, dataGuidNode;

            el = document.elementFromPoint(point.x, point.y);

            if (!el) {
                return null;
            }

            if (!_attachmentArea) {
                _attachmentArea = document.getElementById(ATTACHMENTS_AREA_ID);
            }
            
            if (_attachmentArea.contains && _attachmentArea.contains(el)) {
                node = el;
                while (node) {
                    if (!node.classList) {
                        return null;
                    }

                    if ((node.classList.contains("document") || node.classList.contains("video") || node.classList.contains("audio"))) {
                        json = {
                            "class": node.getAttribute("class"),
                            "data-guid": node.getAttribute("data-guid"),
                            "download-src": node.getAttribute("download-src"),
                            "filename": node.getAttribute("filename"),
                            "filesize": node.getAttribute("filesize"),
                            "isComposing": isComposing()
                        };

                        return JSON.stringify(json);
                    }

                    if (node.classList.contains("image")) {
                        imageNodes = node.getElementsByTagName("IMG");
                        if (imageNodes) {
                            thumbnailData = imageDataFromImageElement(imageNodes[0]);
                            json = {
                                "class": node.getAttribute("class"),
                                "data-guid": node.getAttribute("data-guid"),
                                "download-src": node.getAttribute("download-src"),
                                "filename": node.getAttribute("filename"),
                                "filesize": node.getAttribute("filesize"),
                                "thumbnail": thumbnailData,
                                "isComposing": isComposing()
                            };
                            return JSON.stringify(json);
                        }

                        return null;
                    }

                    node = node.parentNode;
                }
            } else if (el.nodeName == "IMG") {
                node = el;
                if (node.className == CUSTOM_INLINE_IMAGE || node.className == CUSTOM_INLINE_VIDEO || node.className == CUSTOM_INLINE_DOCUMENT || node.className == CUSTOM_INLINE_AUDIO) {
                    while (node != null) {
                        if (node.classList.contains(INLINE_ATTACHMENT_CONTAINER_ID)) {
                            break;
                        } else {
                            node = node.parentNode;
                        }
                    }
                }

                if (!node) {
                    node = el;
                }

                thumbnailData = imageDataFromImageElement(el); //el is always the img node where we get the thumbnail.
                downloadSrcNode = getImageSourceNode(node);
                filenameNode = getFilenameNode(node);
                filesizeNode = getFilesizeNode(node);
                altNode = node.getAttributeNode("alt");
                dataGuidNode = node.getAttributeNode("data-guid");

                json = {
                    "class": node.getAttributeNode("class").value,
                    "download-src": downloadSrcNode ? downloadSrcNode.value : "",
                    "data-guid": dataGuidNode ? dataGuidNode.value : "",
                    "filename": filenameNode ? filenameNode.value : "",
                    "filesize": filesizeNode ? filesizeNode.value : 0,
                    "alt": altNode ? altNode.value : "",
                    "thumbnail": thumbnailData,
                    "isComposing": isComposing()
                };
                return JSON.stringify(json);
            }

            return null;
        },

        imageDataFromImageElement = function(img) {
            var canvas = document.createElement("canvas");
            var context = canvas.getContext("2d");
            var imageDataURL = null;
            canvas.width = img.width;
            canvas.height = img.height;
            context.drawImage(img, 0, 0, img.width, img.height);
            
            try {
                imageDataURL = canvas.toDataURL("image/png");
            } catch(err) {
                // Error caused due to Cross Origin policy
            }
                                 
            return imageDataURL;
        },

        isComposing = function() {
            if (document.querySelector("[contenteditable='true']")) {
                return true;
            } else {
                return false
            }
        },

        getImageSourceNode = function(node) {
            var downloadSrcNode = node.getAttributeNode("download-src");
            if (!downloadSrcNode) {
                downloadSrcNode = node.getAttributeNode("src");
            }
            if (!downloadSrcNode) {
                downloadSrcNode = node.getAttributeNode("data-image-source");
            }
            return downloadSrcNode;
        },

        getFilenameNode = function(node) {
            var filenameNode = node.getAttributeNode("filename");
            if (!filenameNode) {
                filenameNode = node.getAttributeNode("data-filename");
            }
            return filenameNode;
        },

        getFilesizeNode = function(node) {
            var filesizeNode = node.getAttributeNode("filesize");
            if (!filesizeNode) {
                filesizeNode = node.getAttributeNode("data-filesize");
            }
            return filesizeNode;
        },

        attachTouchListener = function() {
            if (!(window.webkit && window.webkit.messageHandlers && window.webkit.messageHandlers.nativeMessageHandler)) {
                return false;
            }

            document.addEventListener("touchstart", function(e) {
                var point = {
                    x: e.touches[0].clientX,
                    y: e.touches[0].clientY
                };

                window.webkit.messageHandlers.nativeMessageHandler.postMessage(webViewDragAndDropHandler.dragItemAtPoint(point));
            }, true);
            
            return true;
        };
    
    // MARK: - Public methods
    // All methods below must return a value to avoid `Fatal error: Bug in WebKit: Received neither result or failure` when evaluated via `WKWebView.evaluateJavaScript()`
    return {
        dragItemAtPoint: dragItemAtPoint,
        attachTouchListener: attachTouchListener
    };
}());
